@@ -1,13 +1,6 @@
 /* Define if you have MIT Kerberos version 4 available.  */
 #undef HAVE_KERBEROS
 
-/* Define if you have GSSAPI with MIT Kerberos version 5 available.  */
-#undef HAVE_GSSAPI
-
-/* Define if GSS_C_NT_HOSTBASED_SERVICE is defined in the gssapi.h
-   header file.  Only relevant when using GSSAPI.  */
-#undef HAVE_GSS_C_NT_HOSTBASED_SERVICE
-
 /* Define if you want CVS to be able to be a remote repository client.  */
 #undef CLIENT_SUPPORT
 
@@ -43,3 +36,15 @@
 /* Define if the diff library should use setmode for binary files.
    FIXME: Why two different macros for setmode?  */
 #undef HAVE_SETMODE
+
+/* Define if you have the crypt function.  */
+#undef HAVE_CRYPT
+
+/* Define if you have the getspnam function.  */
+#undef HAVE_GETSPNAM
+
+/* Define to force lib/regex.c to use malloc instead of alloca.  */
+#undef REGEX_MALLOC
+
+/* Define to force lib/regex.c to define re_comp et al.  */
+#undef _REGEX_RE_COMP
