@@ -37,15 +37,14 @@
 #include "cvs.h"
 
 static char *findslash PROTO((char *start, char *p));
-static int checkout_proc PROTO((int *pargc, char **argv, char *where,
+static int checkout_proc PROTO((int argc, char **argv, char *where,
 		          char *mwhere, char *mfile, int shorten,
 		          int local_specified, char *omodule,
 		          char *msg));
-static int safe_location PROTO((void));
 
 static const char *const checkout_usage[] =
 {
-    "Usage:\n  %s %s [-ANPRcflnps] [-r rev | -D date] [-d dir]\n",
+    "Usage:\n  %s %s [-ANPRcflnps] [-r rev] [-D date] [-d dir]\n",
     "    [-j rev1] [-j rev2] [-k kopt] modules...\n",
     "\t-A\tReset any sticky tags/date/kopts.\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
@@ -60,7 +59,7 @@ static const char *const checkout_usage[] =
     "\t-r rev\tCheck out revision or tag. (implies -P) (is sticky)\n",
     "\t-D date\tCheck out revisions as of date. (implies -P) (is sticky)\n",
     "\t-d dir\tCheck out into dir instead of module name.\n",
-    "\t-k kopt\tUse RCS kopt -k option on checkout.\n",
+    "\t-k kopt\tUse RCS kopt -k option on checkout. (is sticky)\n",
     "\t-j rev\tMerge in changes made between current revision and rev.\n",
     "(Specify the --help global option for a list of other help options)\n",
     NULL
@@ -68,7 +67,7 @@ static const char *const checkout_usage[] =
 
 static const char *const export_usage[] =
 {
-    "Usage: %s %s [-NRfln] [-r rev | -D date] [-d dir] [-k kopt] module...\n",
+    "Usage: %s %s [-NRfln] [-r rev] [-D date] [-d dir] [-k kopt] module...\n",
     "\t-N\tDon't shorten module paths if -d specified.\n",
     "\t-f\tForce a head revision match if tag/date not found.\n",
     "\t-l\tLocal directory only, not recursive\n",
@@ -83,18 +82,19 @@ static const char *const export_usage[] =
 };
 
 static int checkout_prune_dirs;
-static int force_tag_match = 1;
+static int force_tag_match;
 static int pipeout;
 static int aflag;
-static char *options = NULL;
-static char *tag = NULL;
-static int tag_validated = 0;
-static char *date = NULL;
-static char *join_rev1 = NULL;
-static char *join_rev2 = NULL;
-static int join_tags_validated = 0;
-static char *preload_update_dir = NULL;
-static char *history_name = NULL;
+static char *options;
+static char *tag;
+static int tag_validated;
+static char *date;
+static char *join_rev1;
+static char *join_rev2;
+static int join_tags_validated;
+static char *preload_update_dir;
+static char *history_name;
+static enum mtype m_type;
 
 int
 checkout (argc, argv)
@@ -111,7 +111,18 @@ checkout (argc, argv)
     char *where = NULL;
     char *valid_options;
     const char *const *valid_usage;
-    enum mtype m_type;
+
+    /* initialize static options */
+    force_tag_match = 1;
+    if (options)
+    {
+	free (options);
+	options = NULL;
+    }
+    tag = date = join_rev1 = join_rev2 = preload_update_dir = NULL;
+    history_name = NULL;
+    tag_validated = join_tags_validated = 0;
+
 
     /*
      * A smaller subset of options are allowed for the export command, which
@@ -191,7 +202,7 @@ checkout (argc, argv)
 		    shorten = 1;
 		break;
 	    case 's':
-		status = 1;
+		cat = status = 1;
 		break;
 	    case 'f':
 		force_tag_match = 0;
@@ -224,16 +235,16 @@ checkout (argc, argv)
     if (shorten == -1)
 	shorten = 0;
 
-    if ((cat || status) && argc != 0)
+    if (cat && argc != 0)
 	error (1, 0, "-c and -s must not get any arguments");
 
-    if (!(cat || status) && argc == 0)
+    if (!cat && argc == 0)
 	error (1, 0, "must specify at least one module or directory");
 
     if (where && pipeout)
 	error (1, 0, "-d and -p are mutually exclusive");
 
-    if (strcmp (command_name, "export") == 0)
+    if (m_type == EXPORT)
     {
 	if (!tag && !date)
 	    error (1, 0, "must specify a tag or date");
@@ -249,12 +260,12 @@ checkout (argc, argv)
     }
 #endif
 
-    if (!safe_location()) {
+    if (!cat && !pipeout && !safe_location( where )) {
         error(1, 0, "Cannot check out files into the repository itself");
     }
 
 #ifdef CLIENT_SUPPORT
-    if (client_active)
+    if (current_parsed_root->isremote)
     {
 	int expand_modules;
 
@@ -270,7 +281,7 @@ checkout (argc, argv)
            below in !expand_modules), those files (CVS/Checkin.prog
            or CVS/Update.prog) don't get created.  Grrr.  */
 	
-	expand_modules = (!cat && !status && !pipeout
+	expand_modules = (!cat && !pipeout
 			  && supported_request ("expand-modules"));
 	
 	if (expand_modules)
@@ -294,10 +305,10 @@ checkout (argc, argv)
 	    send_arg("-A");
 	if (!shorten)
 	    send_arg("-N");
-	if (checkout_prune_dirs && strcmp (command_name, "export") != 0)
+	if (checkout_prune_dirs && m_type == CHECKOUT)
 	    send_arg("-P");
 	client_prune_dirs = checkout_prune_dirs;
-	if (cat)
+	if (cat && !status)
 	    send_arg("-c");
 	if (where != NULL)
 	    option_with_arg ("-d", where);
@@ -325,19 +336,19 @@ checkout (argc, argv)
 	    client_nonexpanded_setup ();
 	}
 
-	send_to_server (strcmp (command_name, "export") == 0 ?
-                        "export\012" : "co\012",
-                        0);
-
+	send_to_server (m_type == EXPORT ? "export\012" : "co\012", 0);
 	return get_responses_and_close ();
     }
 #endif /* CLIENT_SUPPORT */
 
-    if (cat || status)
+    if (cat)
     {
 	cat_module (status);
 	if (options)
+	{
 	    free (options);
+	    options = NULL;
+	}
 	return (0);
     }
     db = open_module ();
@@ -355,34 +366,44 @@ checkout (argc, argv)
 
     /* If we will be calling history_write, work out the name to pass
        it.  */
-    if (strcmp (command_name, "export") != 0 && !pipeout)
+    if (!pipeout)
     {
-	if (tag && date)
+	if (!date)
+	    history_name = tag;
+	else if (!tag)
+	    history_name = date;
+	else
 	{
 	    history_name = xmalloc (strlen (tag) + strlen (date) + 2);
 	    sprintf (history_name, "%s:%s", tag, date);
 	}
-	else if (tag)
-	    history_name = tag;
-	else
-	    history_name = date;
     }
 
 
     for (i = 0; i < argc; i++)
 	err += do_module (db, argv[i], m_type, "Updating", checkout_proc,
-			  where, shorten, local, run_module_prog,
+			  where, shorten, local, run_module_prog, !pipeout,
 			  (char *) NULL);
     close_module (db);
     if (options)
+    {
 	free (options);
+	options = NULL;
+    }
+    if (history_name != tag && history_name != date && history_name != NULL)
+	free (history_name);
     return (err);
 }
 
-static int
-safe_location ()
+/* FIXME: This is and emptydir_name are in checkout.c for historical
+   reasons, probably want to move them.  */
+
+int
+safe_location (where)
+    char *where;
 {
     char *current;
+    char *where_location;
     char hardpath[PATH_MAX+5];
     size_t hardpath_len;
     int  x;
@@ -392,21 +413,80 @@ safe_location ()
     /* FIXME-arbitrary limit: should be retrying this like xgetwd.
        But how does readlink let us know that the buffer was too small?
        (by returning sizeof hardpath - 1?).  */
-    x = readlink(CVSroot_directory, hardpath, sizeof hardpath - 1);
+    x = readlink(current_parsed_root->directory, hardpath, sizeof hardpath - 1);
 #else
     x = -1;
 #endif
     if (x == -1)
     {
-        strcpy(hardpath, CVSroot_directory);
+        strcpy(hardpath, current_parsed_root->directory);
     }
     else
     {
         hardpath[x] = '\0';
     }
+
+    /* set current - even if where is set we'll need to cd back... */
     current = xgetwd ();
     if (current == NULL)
 	error (1, errno, "could not get working directory");
+
+    /* if where is set, set current to where, where - last_component( where ),
+     * or fail, depending on whether the directories exist or not.
+     */
+    if( where != NULL )
+    {
+	if( chdir( where ) != -1 )
+	{
+	    /* where */
+	    where_location = xgetwd();
+	    if( where_location == NULL )
+		error( 1, errno, "could not get working directory" );
+
+	    if( chdir( current ) == -1 )
+		error( 1, errno, "could not change directory to `%s'", current );
+
+	    free( current );
+	    current = where_location;
+        }
+	else if( errno == ENOENT )
+	{
+	    if ( last_component( where ) != where )
+	    {
+		/* where - last_component( where ) */
+		char *parent;
+
+		/* strip the last_component */
+		where_location = strdup( where );
+		parent = last_component( where_location );
+		parent[-1] = '\0';
+
+		if( chdir( where_location ) != -1 )
+		{
+		    where_location = xgetwd();
+		    if( where_location == NULL )
+			error( 1, errno, "could not get working directory (nominally `%s')", where_location );
+
+		    if( chdir( current ) == -1 )
+			error( 1, errno, "could not change directory to `%s'", current );
+
+		    free( current );
+		    current = where_location;
+		}
+		else
+		    /* fail */
+		    error( 1, errno, "could not change directory to requested checkout directory `%s'", where_location );
+	    }
+	    /* else: ERRNO == ENOENT & last_component(where) == where
+	     * for example, 'cvs co -d newdir module', where newdir hasn't
+	     * been created yet, so leave current set to '.' and check that
+	     */
+	}
+	else
+	    /* fail */
+	    error( 1, errno, "could not change directory to requested checkout directory `%s'", where );
+    }
+
     hardpath_len = strlen (hardpath);
     if (strlen (current) >= hardpath_len
 	&& strncmp (current, hardpath, hardpath_len) == 0)
@@ -455,15 +535,20 @@ build_one_dir (repository, dirpath, sticky)
 {
     FILE *fp;
 
-    if (!isfile (CVSADM) && strcmp (command_name, "export") != 0)
+    if (isfile (CVSADM))
+    {
+	if (m_type == EXPORT)
+	    error (1, 0, "cannot export into a working directory");
+    }
+    else if (m_type == CHECKOUT)
     {
 	/* I suspect that this check could be omitted.  */
 	if (!isdir (repository))
 	    error (1, 0, "there is no repository %s", repository);
 
 	if (Create_Admin (".", dirpath, repository,
-			  sticky ? (char *) NULL : tag,
-			  sticky ? (char *) NULL : date,
+			  sticky ? tag : (char *) NULL,
+			  sticky ? date : (char *) NULL,
 
 			  /* FIXME?  This is a guess.  If it is important
 			     for nonbranch to be set correctly here I
@@ -471,7 +556,7 @@ build_one_dir (repository, dirpath, sticky)
 			     then rewrite it later via WriteTag, once
 			     we've had a chance to call RCS_nodeisbranch
 			     on each file.  */
-			  0, 1))
+			  0, 1, 1))
 	    return;
 
 	if (!noexec)
@@ -492,9 +577,9 @@ build_one_dir (repository, dirpath, sticky)
  */
 /* ARGSUSED */
 static int
-checkout_proc (pargc, argv, where_orig, mwhere, mfile, shorten,
+checkout_proc (argc, argv, where_orig, mwhere, mfile, shorten,
 	       local_specified, omodule, msg)
-    int *pargc;
+    int argc;
     char **argv;
     char *where_orig;
     char *mwhere;
@@ -504,6 +589,7 @@ checkout_proc (pargc, argv, where_orig, mwhere, mfile, shorten,
     char *omodule;
     char *msg;
 {
+    char *myargv[2];
     int err = 0;
     int which;
     char *cp;
@@ -524,11 +610,11 @@ checkout_proc (pargc, argv, where_orig, mwhere, mfile, shorten,
     /* Set up the repository (maybe) for the bottom directory.
        Allocate more space than we need so we don't need to keep
        reallocating this string. */
-    repository = xmalloc (strlen (CVSroot_directory)
+    repository = xmalloc (strlen (current_parsed_root->directory)
 			  + strlen (argv[0])
 			  + (mfile == NULL ? 0 : strlen (mfile))
 			  + 10);
-    (void) sprintf (repository, "%s/%s", CVSroot_directory, argv[0]);
+    (void) sprintf (repository, "%s/%s", current_parsed_root->directory, argv[0]);
     Sanitize_Repository_Name (repository);
 
 
@@ -665,26 +751,10 @@ checkout_proc (pargc, argv, where_orig, mwhere, mfile, shorten,
 	{
 	    /* It's a file, which means we have to screw around with
                argv. */
-
-	    int i;
-
-
-	    /* Paranoia check. */
-	    
-	    if (*pargc > 1)
-	    {
-		error (0, 0, "checkout_proc: trashing argv elements!");
-		for (i = 1; i < *pargc; i++)
-		{
-		    error (0, 0, "checkout_proc: argv[%d] `%s'",
-			   i, argv[i]);
-		}
-	    }
-
-	    for (i = 1; i < *pargc; i++)
-		free (argv[i]);
-	    argv[1] = xstrdup (mfile);
-	    (*pargc) = 2;
+	    myargv[0] = argv[0];
+	    myargv[1] = mfile;
+	    argc = 2;
+	    argv = myargv;
 	}
 	free (path);
     }
@@ -720,11 +790,11 @@ checkout_proc (pargc, argv, where_orig, mwhere, mfile, shorten,
 	struct dir_to_build *head;
 	char *reposcopy;
 
-	if (strncmp (repository, CVSroot_directory,
-		     strlen (CVSroot_directory)) != 0)
+	if (strncmp (repository, current_parsed_root->directory,
+		     strlen (current_parsed_root->directory)) != 0)
 	    error (1, 0, "\
 internal error: %s doesn't start with %s in checkout_proc",
-		   repository, CVSroot_directory);
+		   repository, current_parsed_root->directory);
 
 	/* We always create at least one directory, which corresponds to
 	   the entire strings for WHERE and REPOSITORY.  */
@@ -745,7 +815,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 	   NT, &c, if the user specifies '\'.  Likewise for the call
 	   to findslash.  */
 	cp = where + strlen (where);
-	while (1)
+	while (cp > where)
 	{
 	    struct dir_to_build *new;
 
@@ -761,7 +831,7 @@ internal error: %s doesn't start with %s in checkout_proc",
                last path element we create should be the top-level
                directory. */
 
-	    if (cp - where)
+	    if (cp > where)
 	    {
 		strncpy (new->dirpath, where, cp - where);
 		new->dirpath[cp - where] = '\0';
@@ -769,7 +839,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 	    else
 	    {
 		/* where should always be at least one character long. */
-		assert (strlen (where));
+		assert (where[0] != '\0');
 		strcpy (new->dirpath, "/");
 	    }
 	    new->next = head;
@@ -809,7 +879,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 		 bar   -> Emptydir   (generated dir -- not in repos)
 		 baz   -> quux       (finally!) */
 
-	    if (strcmp (reposcopy, CVSroot_directory) == 0)
+	    if (strcmp (reposcopy, current_parsed_root->directory) == 0)
 	    {
 		/* We can't walk up past CVSROOT.  Instead, the
                    repository should be Emptydir. */
@@ -817,54 +887,30 @@ internal error: %s doesn't start with %s in checkout_proc",
 	    }
 	    else
 	    {
-		if ((where_orig != NULL)
-		    && (strcmp (new->dirpath, where_orig) == 0))
-		{
-		    /* It's the case that the user specified a
-		     * destination directory with the "-d" flag.  The
-		     * repository in this directory should be "."
-		     * since the user's command is equivalent to:
-		     *
-		     *   cd <dir>; cvs co blah   */
-
-		    strcpy (reposcopy, CVSroot_directory);
-		    goto allocate_repos;
-		}
-		else if (mwhere != NULL)
-		{
-		    /* This is a generated directory, so point to
-                       CVSNULLREPOS. */
-
-		    new->repository = emptydir_name ();
-		}
-		else
-		{
-		    /* It's a directory in the repository! */
+		/* It's a directory in the repository! */
 		    
-		    char *rp = strrchr (reposcopy, '/');
+		char *rp;
 		    
-		    /* We'll always be below CVSROOT, but check for
-		       paranoia's sake. */
-		    if (rp == NULL)
-			error (1, 0,
-			       "internal error: %s doesn't contain a slash",
-			       reposcopy);
+		/* We'll always be below CVSROOT, but check for
+		   paranoia's sake. */
+		rp = strrchr (reposcopy, '/');
+		if (rp == NULL)
+		    error (1, 0,
+			   "internal error: %s doesn't contain a slash",
+			   reposcopy);
 			   
-		    *rp = '\0';
-		
-		allocate_repos:
-		    new->repository = xmalloc (strlen (reposcopy) + 5);
-		    (void) strcpy (new->repository, reposcopy);
+		*rp = '\0';
+		new->repository = xmalloc (strlen (reposcopy) + 5);
+		(void) strcpy (new->repository, reposcopy);
 		    
-		    if (strcmp (reposcopy, CVSroot_directory) == 0)
-		    {
-			/* Special case -- the repository name needs
-			   to be "/path/to/repos/." (the trailing dot
-			   is important).  We might be able to get rid
-			   of this after the we check out the other
-			   code that handles repository names. */
-			(void) strcat (new->repository, "/.");
-		    }
+		if (strcmp (reposcopy, current_parsed_root->directory) == 0)
+		{
+		    /* Special case -- the repository name needs
+		       to be "/path/to/repos/." (the trailing dot
+		       is important).  We might be able to get rid
+		       of this after the we check out the other
+		       code that handles repository names. */
+		    (void) strcat (new->repository, "/.");
 		}
 	    }
 	}
@@ -876,7 +922,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 	    int where_is_absolute = isabsolute (where);
 	    
 	    /* The top-level CVSADM directory should always be
-	       CVSroot_directory.  Create it, but only if WHERE is
+	       current_parsed_root->directory.  Create it, but only if WHERE is
 	       relative.  If WHERE is absolute, our current directory
 	       may not have a thing to do with where the sources are
 	       being checked out.  If it does, build_dirs_and_chdir
@@ -886,11 +932,11 @@ internal error: %s doesn't start with %s in checkout_proc",
 	       build_one_dir whenever the -d command option was specified
 	       to checkout.  */
 
-	    if (! where_is_absolute && top_level_admin)
+	    if (! where_is_absolute && top_level_admin && m_type == CHECKOUT)
 	    {
 		/* It may be argued that we shouldn't set any sticky
 		   bits for the top-level repository.  FIXME?  */
-		build_one_dir (CVSroot_directory, ".", *pargc <= 1);
+		build_one_dir (current_parsed_root->directory, ".", argc <= 1);
 
 #ifdef SERVER_SUPPORT
 		/* We _always_ want to have a top-level admin
@@ -902,7 +948,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 		   will be ignored on the client side.  */
 
 		if (server_active)
-		    server_clear_entstat (".", CVSroot_directory);
+		    server_clear_entstat (".", current_parsed_root->directory);
 #endif
 	    }
 
@@ -912,7 +958,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 	       contain a CVS subdir yet, but all the others contain
 	       CVS and Entries.Static files */
 
-	    if (build_dirs_and_chdir (head, *pargc <= 1) != 0)
+	    if (build_dirs_and_chdir (head, argc <= 1) != 0)
 	    {
 		error (0, 0, "ignoring module %s", omodule);
 		err = 1;
@@ -925,14 +971,15 @@ internal error: %s doesn't start with %s in checkout_proc",
 	{
 	    FILE *fp;
 
-	    if (!noexec && *pargc > 1)
+	    if (!noexec && argc > 1)
 	    {
 		/* I'm not sure whether this check is redundant.  */
 		if (!isdir (repository))
 		    error (1, 0, "there is no repository %s", repository);
 
 		Create_Admin (".", preload_update_dir, repository,
-			      (char *) NULL, (char *) NULL, 0, 0);
+			      (char *) NULL, (char *) NULL, 0, 0,
+			      m_type == CHECKOUT);
 		fp = open_file (CVSADM_ENTSTAT, "w+");
 		if (fclose(fp) == EOF)
 		    error(1, errno, "cannot close %s", CVSADM_ENTSTAT);
@@ -955,13 +1002,16 @@ internal error: %s doesn't start with %s in checkout_proc",
 				 then rewrite it later via WriteTag, once
 				 we've had a chance to call RCS_nodeisbranch
 				 on each file.  */
-			      0, 0);
+			      0, 0, m_type == CHECKOUT);
 	    }
 	}
 	else
 	{
 	    char *repos;
 
+	    if (m_type == EXPORT)
+		error (1, 0, "cannot export into working directory");
+
 	    /* get the contents of the previously existing repository */
 	    repos = Name_Repository ((char *) NULL, preload_update_dir);
 	    if (fncmp (repository, repos) != 0)
@@ -993,7 +1043,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 	which = W_REPOS;
 	if (tag != NULL && !tag_validated)
 	{
-	    tag_check_valid (tag, *pargc - 1, argv + 1, 0, aflag, NULL);
+	    tag_check_valid (tag, argc - 1, argv + 1, 0, aflag, NULL);
 	    tag_validated = 1;
 	}
     }
@@ -1002,7 +1052,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 	which = W_LOCAL | W_REPOS;
 	if (tag != NULL && !tag_validated)
 	{
-	    tag_check_valid (tag, *pargc - 1, argv + 1, 0, aflag,
+	    tag_check_valid (tag, argc - 1, argv + 1, 0, aflag,
 			     repository);
 	    tag_validated = 1;
 	}
@@ -1014,10 +1064,10 @@ internal error: %s doesn't start with %s in checkout_proc",
     if (! join_tags_validated)
     {
         if (join_rev1 != NULL)
-	    tag_check_valid_join (join_rev1, *pargc - 1, argv + 1, 0, aflag,
+	    tag_check_valid_join (join_rev1, argc - 1, argv + 1, 0, aflag,
 				  repository);
 	if (join_rev2 != NULL)
-	    tag_check_valid_join (join_rev2, *pargc - 1, argv + 1, 0, aflag,
+	    tag_check_valid_join (join_rev2, argc - 1, argv + 1, 0, aflag,
 				  repository);
 	join_tags_validated = 1;
     }
@@ -1027,19 +1077,16 @@ internal error: %s doesn't start with %s in checkout_proc",
      * update recursion processor.  We will be recursive unless either local
      * only was specified, or we were passed arguments
      */
-    if (!(local_specified || *pargc > 1))
+    if (!(local_specified || argc > 1))
     {
-	if (strcmp (command_name, "export") != 0 && !pipeout)
-	    history_write ('O', preload_update_dir, history_name, where,
-			   repository);
-	else if (strcmp (command_name, "export") == 0 && !pipeout)
-	    history_write ('E', preload_update_dir, tag ? tag : date, where,
-			   repository);
+	if (!pipeout)
+	    history_write (m_type == CHECKOUT ? 'O' : 'E', preload_update_dir,
+			   history_name, where, repository);
 	err += do_update (0, (char **) NULL, options, tag, date,
 			  force_tag_match, 0 /* !local */ ,
 			  1 /* update -d */ , aflag, checkout_prune_dirs,
 			  pipeout, which, join_rev1, join_rev2,
-			  preload_update_dir);
+			  preload_update_dir, m_type == CHECKOUT);
 	goto out;
     }
 
@@ -1050,7 +1097,7 @@ internal error: %s doesn't start with %s in checkout_proc",
 
 	/* we are only doing files, so register them */
 	entries = Entries_Open (0, NULL);
-	for (i = 1; i < *pargc; i++)
+	for (i = 1; i < argc; i++)
 	{
 	    char *line;
 	    Vers_TS *vers;
@@ -1087,15 +1134,15 @@ internal error: %s doesn't start with %s in checkout_proc",
     }
 
     /* Don't log "export", just regular "checkouts" */
-    if (strcmp (command_name, "export") != 0 && !pipeout)
+    if (m_type == CHECKOUT && !pipeout)
 	history_write ('O', preload_update_dir, history_name, where,
 		       repository);
 
     /* go ahead and call update now that everything is set */
-    err += do_update (*pargc - 1, argv + 1, options, tag, date,
+    err += do_update (argc - 1, argv + 1, options, tag, date,
 		      force_tag_match, local_specified, 1 /* update -d */,
 		      aflag, checkout_prune_dirs, pipeout, which, join_rev1,
-		      join_rev2, preload_update_dir);
+		      join_rev2, preload_update_dir, m_type == CHECKOUT);
 out:
     free (preload_update_dir);
     preload_update_dir = oldupdate;
@@ -1109,15 +1156,13 @@ findslash (start, p)
     char *start;
     char *p;
 {
-    while (p >= start && *p != '/')
-	p--;
-    /* FIXME: indexing off the start of the array like this is *NOT*
-       OK according to ANSI, and will break some of the time on certain
-       segmented architectures.  */
-    if (p < start)
-	return (NULL);
-    else
-	return (p);
+    for (;;)
+    {
+	if (*p == '/') return p;
+	if (p == start) break;
+	--p;
+    }
+    return NULL;
 }
 
 /* Return a newly malloc'd string containing a pathname for CVSNULLREPOS,
@@ -1129,11 +1174,11 @@ emptydir_name ()
 {
     char *repository;
 
-    repository = xmalloc (strlen (CVSroot_directory) 
+    repository = xmalloc (strlen (current_parsed_root->directory) 
 			  + sizeof (CVSROOTADM)
 			  + sizeof (CVSNULLREPOS)
-			  + 10);
-    (void) sprintf (repository, "%s/%s/%s", CVSroot_directory,
+			  + 3);
+    (void) sprintf (repository, "%s/%s/%s", current_parsed_root->directory,
 		    CVSROOTADM, CVSNULLREPOS);
     if (!isfile (repository))
     {
@@ -1185,5 +1230,14 @@ build_dirs_and_chdir (dirs, sticky)
     }
 
  out:
+    while (dirs != NULL)
+    {
+	if (dirs->repository != NULL)
+	    free (dirs->repository);
+	nextdir = dirs->next;
+	free (dirs->dirpath);
+	free (dirs);
+	dirs = nextdir;
+    }
     return retval;
 }
