@@ -272,7 +272,7 @@ expand_variable (name, file, line)
     int line;
 {
     if (strcmp (name, CVSROOT_ENV) == 0)
-	return CVSroot_original;
+	return current_parsed_root->original;
     else if (strcmp (name, "RCSBIN") == 0)
     {
 	error (0, 0, "RCSBIN internal variable is no longer supported");
