@@ -17,6 +17,7 @@
    definitions under operating systems (like, say, Windows NT) with different
    file system semantics.  */
 
+#include <assert.h>
 #include "cvs.h"
 
 static int deep_remove_dir PROTO((const char *path));
@@ -34,12 +35,8 @@ copy_file (from, to)
     int fdin, fdout;
 
     if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> copy(%s,%s)\n",
-			(server_active) ? 'S' : ' ', from, to);
-#else
-	(void) fprintf (stderr, "-> copy(%s,%s)\n", from, to);
-#endif
+	(void) fprintf (stderr, "%s-> copy(%s,%s)\n",
+			CLIENT_SERVER_STR, from, to);
     if (noexec)
 	return;
 
@@ -55,9 +52,13 @@ copy_file (from, to)
 
     if (isdevice (from))
     {
+#if defined(HAVE_MKNOD) && defined(HAVE_ST_RDEV)
 	if (stat (from, &sb) < 0)
 	    error (1, errno, "cannot stat %s", from);
 	mknod (to, sb.st_mode, sb.st_rdev);
+#else
+	error (1, 0, "cannot copy device files on this system (%s)", from);
+#endif
     }
     else
     {
@@ -330,8 +331,7 @@ mkdir_if_needed (name)
 {
     if (mkdir (name, 0777) < 0)
     {
-	if (!(errno == EEXIST
-	      || (errno == EACCES && isdir (name))))
+	if (errno != EEXIST && !isdir (name))
 	    error (1, errno, "cannot make directory %s", name);
 	return 1;
     }
@@ -377,14 +377,9 @@ xchmod (fname, writable)
     }
 
     if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> chmod(%s,%o)\n",
-			(server_active) ? 'S' : ' ', fname,
+	(void) fprintf (stderr, "%s-> chmod(%s,%o)\n",
+			CLIENT_SERVER_STR, fname,
 			(unsigned int) mode);
-#else
-	(void) fprintf (stderr, "-> chmod(%s,%o)\n", fname,
-			(unsigned int) mode);
-#endif
     if (noexec)
 	return;
 
@@ -401,12 +396,8 @@ rename_file (from, to)
     const char *to;
 {
     if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> rename(%s,%s)\n",
-			(server_active) ? 'S' : ' ', from, to);
-#else
-	(void) fprintf (stderr, "-> rename(%s,%s)\n", from, to);
-#endif
+	(void) fprintf (stderr, "%s-> rename(%s,%s)\n",
+			CLIENT_SERVER_STR, from, to);
     if (noexec)
 	return;
 
@@ -422,16 +413,12 @@ unlink_file (f)
     const char *f;
 {
     if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> unlink(%s)\n",
-			(server_active) ? 'S' : ' ', f);
-#else
-	(void) fprintf (stderr, "-> unlink(%s)\n", f);
-#endif
+	(void) fprintf (stderr, "%s-> unlink_file(%s)\n",
+			CLIENT_SERVER_STR, f);
     if (noexec)
 	return (0);
 
-    return (unlink (f));
+    return (CVS_UNLINK (f));
 }
 
 /*
@@ -477,7 +464,7 @@ unlink_file_dir (f)
     else if (S_ISDIR (sb.st_mode))
 	return deep_remove_dir (f);
 
-    return unlink (f);
+    return CVS_UNLINK (f);
 }
 
 /* Remove a directory and everything it contains.  Returns 0 for
@@ -500,13 +487,14 @@ deep_remove_dir (path)
 	       returns 87).  */
 	    || (ENOTEMPTY == 17 && EEXIST == 17 && errno == 87))
 	{
-	    if ((dirp = opendir (path)) == NULL)
+	    if ((dirp = CVS_OPENDIR (path)) == NULL)
 		/* If unable to open the directory return
 		 * an error
 		 */
 		return -1;
 
-	    while ((dp = readdir (dirp)) != NULL)
+	    errno = 0;
+	    while ((dp = CVS_READDIR (dirp)) != NULL)
 	    {
 		char *buf;
 
@@ -524,23 +512,32 @@ deep_remove_dir (path)
 		{
 		    if (deep_remove_dir(buf))
 		    {
-			closedir(dirp);
+			CVS_CLOSEDIR(dirp);
 			free (buf);
 			return -1;
 		    }
 		}
 		else
 		{
-		    if (unlink (buf) != 0)
+		    if (CVS_UNLINK (buf) != 0)
 		    {
-			closedir(dirp);
+			CVS_CLOSEDIR(dirp);
 			free (buf);
 			return -1;
 		    }
 		}
 		free (buf);
+
+		errno = 0;
 	    }
-	    closedir (dirp);
+	    if (errno != 0)
+	    {
+		int save_errno = errno;
+		CVS_CLOSEDIR (dirp);
+		errno = save_errno;
+		return -1;
+	    }
+	    CVS_CLOSEDIR (dirp);
 	    return rmdir (path);
 	}
 	else
@@ -627,10 +624,15 @@ xcmp (file1, file2)
        numbers match. */
     if (S_ISBLK (sb1.st_mode) || S_ISCHR (sb1.st_mode))
     {
+#ifdef HAVE_ST_RDEV
 	if (sb1.st_rdev == sb2.st_rdev)
 	    return 0;
 	else
 	    return 1;
+#else
+	error (1, 0, "cannot compare device files on this system (%s and %s)",
+	       file1, file2);
+#endif
     }
 
     if ((fd1 = open (file1, O_RDONLY)) < 0)
@@ -682,53 +684,167 @@ xcmp (file1, file2)
 }
 
 /* Generate a unique temporary filename.  Returns a pointer to a newly
-   malloc'd string containing the name.  Returns successfully or not at
-   all.  */
-/* There are at least three functions for generating temporary
-   filenames.  We use tempnam (SVID 3) if possible, else mktemp (BSD
-   4.3), and as last resort tmpnam (POSIX). Reason is that tempnam and
-   mktemp both allow to specify the directory in which the temporary
-   file will be created.  */
-#ifdef HAVE_TEMPNAM
+ * malloc'd string containing the name.  Returns successfully or not at
+ * all.
+ *
+ *     THIS FUNCTION IS DEPRECATED!!!  USE cvs_temp_file INSTEAD!!!
+ *
+ * and yes, I know about the way the rcs commands use temp files.  I think
+ * they should be converted too but I don't have time to look into it right
+ * now.
+ */
 char *
 cvs_temp_name ()
 {
-    char *retval;
+    char *fn;
+    FILE *fp;
+
+    fp = cvs_temp_file (&fn);
+    if (fp == NULL)
+	error (1, errno, "Failed to create temporary file");
+    if (fclose (fp) == EOF)
+	error (0, errno, "Failed to close temporary file %s", fn);
+    return fn;
+}
+
+/* Generate a unique temporary filename and return an open file stream
+ * to the truncated file by that name
+ *
+ *  INPUTS
+ *	filename	where to place the pointer to the newly allocated file
+ *   			name string
+ *
+ *  OUTPUTS
+ *	filename	dereferenced, will point to the newly allocated file
+ *			name string.  This value is undefined if the function
+ *			returns an error.
+ *
+ *  RETURNS
+ *	An open file pointer to a read/write mode empty temporary file with the
+ *	unique file name or NULL on failure.
+ *
+ *  ERRORS
+ *	on error, errno will be set to some value either by CVS_FOPEN or
+ *	whatever system function is called to generate the temporary file name
+ */
+/* There are at least four functions for generating temporary
+ * filenames.  We use mkstemp (BSD 4.3) if possible, else tempnam (SVID 3),
+ * else mktemp (BSD 4.3), and as last resort tmpnam (POSIX).  Reason is that
+ * mkstemp, tempnam, and mktemp both allow to specify the directory in which
+ * the temporary file will be created.
+ *
+ * And the _correct_ way to use the deprecated functions probably involves
+ * opening file descriptors using O_EXCL & O_CREAT and even doing the annoying
+ * NFS locking thing, but until I hear of more problems, I'm not going to
+ * bother.
+ */
+FILE *cvs_temp_file (filename)
+    char **filename;
+{
+    char *fn;
+    FILE *fp;
+
+    /* FIXME - I'd like to be returning NULL here in noexec mode, but I think
+     * some of the rcs & diff functions which rely on a temp file run in
+     * noexec mode too.
+     */
+
+    assert (filename != NULL);
+
+#ifdef HAVE_MKSTEMP
+
+    {
+    int fd;
+
+    fn = xmalloc (strlen (Tmpdir) + 11);
+    sprintf (fn, "%s/%s", Tmpdir, "cvsXXXXXX" );
+    fd = mkstemp (fn);
+
+    /* a NULL return will be interpreted by callers as an error and
+     * errno should still be set
+     */
+    if (fd == -1) fp = NULL;
+    else if ((fp = CVS_FDOPEN (fd, "w+")) == NULL)
+    {
+	/* attempt to close and unlink the file since mkstemp returned sucessfully and
+	 * we believe it's been created and opened
+	 */
+ 	int save_errno = errno;
+	if (close (fd))
+	    error (0, errno, "Failed to close temporary file %s", fn);
+	if (CVS_UNLINK (fn))
+	    error (0, errno, "Failed to unlink temporary file %s", fn);
+	errno = save_errno;
+    }
+
+    if (fp == NULL) free (fn);
+    /* mkstemp is defined to open mode 0600 using glibc 2.0.7+ */
+    /* FIXME - configure can probably tell us which version of glibc we are
+     * linking to and not chmod for 2.0.7+
+     */
+    else chmod (fn, 0600);
+
+    }
+
+#elif HAVE_TEMPNAM
+
+    /* tempnam has been deprecated due to under-specification */
+
+    fn = tempnam (Tmpdir, "cvs");
+    if (fn == NULL) fp = NULL;
+    else if ((fp = CVS_FOPEN (fn, "w+")) == NULL) free (fn);
+    else chmod (fn, 0600);
 
-    retval = tempnam (Tmpdir, "cvs");
-    if (retval == NULL)
-	error (1, errno, "cannot generate temporary filename");
     /* tempnam returns a pointer to a newly malloc'd string, so there's
-       no need for a xstrdup  */
-    return retval;
-}
-#else
-char *
-cvs_temp_name ()
-{
-#  ifdef HAVE_MKTEMP
-    char *value;
-    char *retval;
+     * no need for a xstrdup
+     */
 
-    value = xmalloc (strlen (Tmpdir) + 40);
-    sprintf (value, "%s/%s", Tmpdir, "cvsXXXXXX" );
-    retval = mktemp (value);
+#elif HAVE_MKTEMP
 
-    if (retval == NULL)
-	error (1, errno, "cannot generate temporary filename");
-    return value;
-#  else
-    char value[L_tmpnam + 1];
-    char *retval;
+    /* mktemp has been deprecated due to the BSD 4.3 specification specifying
+     * that XXXXXX will be replaced by a PID and a letter, creating only 26
+     * possibilities, a security risk, and a race condition.
+     */
+
+    {
+    char *ifn;
+
+    ifn = xmalloc (strlen (Tmpdir) + 11);
+    sprintf (ifn, "%s/%s", Tmpdir, "cvsXXXXXX" );
+    fn = mktemp (ifn);
+
+    if (fn == NULL) fp = NULL;
+    else fp = CVS_FOPEN (fn, "w+");
+
+    if (fp == NULL) free (ifn);
+    else chmod (fn, 0600);
+
+    }
+
+#else	/* use tmpnam if all else fails */
+
+    /* tmpnam is deprecated */
+
+    {
+    char ifn[L_tmpnam + 1];
+
+    fn = tmpnam (ifn);
+
+    if (fn == NULL) fp = NULL;
+    else if ((fp = CVS_FOPEN (ifn, "w+")) != NULL)
+    {
+	fn = xstrdup (ifn);
+	chmod (fn, 0600);
+    }
+
+    }
 
-    retval = tmpnam (value);
-    if (retval == NULL)
-	error (1, errno, "cannot generate temporary filename");
-    return xstrdup (value);
-#  endif
-}
 #endif
-
+
+    *filename = fn;
+    return fp;
+}
+
 /* Return non-zero iff FILENAME is absolute.
    Trivial under Unix, but more complicated under other systems.  */
 int
@@ -817,13 +933,17 @@ char *
 get_homedir ()
 {
     static char *home = NULL;
-    char *env = getenv ("HOME");
+    char *env;
     struct passwd *pw;
 
     if (home != NULL)
 	return home;
 
-    if (env)
+    if (
+#ifdef SERVER_SUPPORT
+	!server_active &&
+#endif
+	(env = getenv ("HOME")) != NULL)
 	home = env;
     else if ((pw = (struct passwd *) getpwuid (getuid ()))
 	     && pw->pw_dir)
@@ -930,7 +1050,7 @@ fopen_case (name, mode, fp, pathp)
 	}
     }
     errno = 0;
-    while ((dp = readdir (dirp)) != NULL)
+    while ((dp = CVS_READDIR (dirp)) != NULL)
     {
 	if (cvs_casecmp (dp->d_name, fname) == 0)
 	{
@@ -942,7 +1062,7 @@ fopen_case (name, mode, fp, pathp)
     }
     if (errno != 0)
 	error (1, errno, "cannot read directory %s", dir);
-    closedir (dirp);
+    CVS_CLOSEDIR (dirp);
 
     if (found_name == NULL)
     {
