@@ -1,8 +1,19 @@
+/* This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2, or (at your option)
+   any later version.
+
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.  */
+
 /*
  * .cvsignore file support contributed by David G. Grubbs <dgg@odi.com>
  */
 
 #include "cvs.h"
+#include "getline.h"
 
 /*
  * Ignore file section.
@@ -43,7 +54,6 @@ void
 ign_setup ()
 {
     char *home_dir;
-    char file[PATH_MAX];
     char *tmp;
 
     ign_inhibit_server = 0;
@@ -58,21 +68,31 @@ ign_setup ()
        processing, and only if !ign_inhibit_server), letting the server
        know about the files and letting it decide whether to ignore
        them based on CVSROOOTADM_IGNORE.  */
-    if (!client_active)
+    if (!current_parsed_root->isremote)
 #endif
     {
+	char *file = xmalloc (strlen (current_parsed_root->directory) + sizeof (CVSROOTADM)
+			      + sizeof (CVSROOTADM_IGNORE) + 10);
 	/* Then add entries found in repository, if it exists */
-	(void) sprintf (file, "%s/%s/%s", CVSroot_directory,
+	(void) sprintf (file, "%s/%s/%s", current_parsed_root->directory,
 			CVSROOTADM, CVSROOTADM_IGNORE);
 	ign_add_file (file, 0);
+	free (file);
     }
 
     /* Then add entries found in home dir, (if user has one) and file exists */
     home_dir = get_homedir ();
+    /* If we can't find a home directory, ignore ~/.cvsignore.  This may
+       make tracking down problems a bit of a pain, but on the other
+       hand it might be obnoxious to complain when CVS will function
+       just fine without .cvsignore (and many users won't even know what
+       .cvsignore is).  */
     if (home_dir)
     {
+	char *file = xmalloc (strlen (home_dir) + sizeof (CVSDOTIGNORE) + 10);
 	(void) sprintf (file, "%s/%s", home_dir, CVSDOTIGNORE);
 	ign_add_file (file, 0);
+	free (file);
     }
 
     /* Then add entries found in CVSIGNORE environment variable. */
@@ -92,7 +112,8 @@ ign_add_file (file, hold)
     int hold;
 {
     FILE *fp;
-    char line[1024];
+    char *line = NULL;
+    size_t line_allocated = 0;
 
     /* restore the saved list (if any) */
     if (s_ign_list != NULL)
@@ -136,10 +157,13 @@ ign_add_file (file, hold)
 	    error (0, errno, "cannot open %s", file);
 	return;
     }
-    while (fgets (line, sizeof (line), fp))
+    while (getline (&line, &line_allocated, fp) >= 0)
 	ign_add (line, hold);
+    if (ferror (fp))
+	error (0, errno, "cannot read %s", file);
     if (fclose (fp) < 0)
 	error (0, errno, "cannot close %s", file);
+    free (line);
 }
 
 /* Parse a line of space-separated wildcards and add them to the list. */
@@ -157,7 +181,7 @@ ign_add (ign, hold)
 	char save;
 
 	/* ignore whitespace before the token */
-	if (isspace (*ign))
+	if (isspace ((unsigned char) *ign))
 	    continue;
 
 	/*
@@ -165,7 +189,8 @@ ign_add (ign, hold)
 	 * (saving it if necessary).  We also catch * as a special case in a
 	 * global ignore file as an optimization
 	 */
-	if ((!*(ign+1) || isspace (*(ign+1))) && (*ign == '!' || *ign == '*'))
+	if ((!*(ign+1) || isspace ((unsigned char) *(ign+1)))
+	    && (*ign == '!' || *ign == '*'))
 	{
 	    if (!hold)
 	    {
@@ -214,7 +239,7 @@ ign_add (ign, hold)
 	}
 
 	/* find the end of this token */
-	for (mark = ign; *mark && !isspace (*mark); mark++)
+	for (mark = ign; *mark && !isspace ((unsigned char) *mark); mark++)
 	     /* do nothing */ ;
 
 	save = *mark;
@@ -231,8 +256,9 @@ ign_add (ign, hold)
     }
 }
 
-/* Set to 1 if ignore file patterns should be matched in a case-insensitive
-   fashion.  */
+/* Set to 1 if filenames should be matched in a case-insensitive
+   fashion.  Note that, contrary to the name and placement in ignore.c,
+   this is no longer just for ignore patterns.  */
 int ign_case;
 
 /* Return 1 if the given filename should be ignored by update or import. */
@@ -249,7 +275,9 @@ ign_name (name)
     {
 	/* We do a case-insensitive match by calling fnmatch on copies of
 	   the pattern and the name which have been converted to
-	   lowercase.  */
+	   lowercase.  FIXME: would be much cleaner to just unify this
+	   with the other case-insensitive fnmatch stuff (FOLD_FN_CHAR
+	   in lib/fnmatch.c; os2_fnmatch in emx/system.c).  */
 	char *name_lower;
 	char *pat_lower;
 	char *p;
@@ -262,7 +290,7 @@ ign_name (name)
 	    pat_lower = xstrdup (*cpp++);
 	    for (p = pat_lower; *p != '\0'; ++p)
 		*p = tolower (*p);
-	    if (fnmatch (pat_lower, name_lower, 0) == 0)
+	    if (CVS_FNMATCH (pat_lower, name_lower, 0) == 0)
 		goto matched;
 	    free (pat_lower);
 	}
@@ -276,55 +304,58 @@ ign_name (name)
     else
     {
 	while (*cpp)
-	    if (fnmatch (*cpp++, name, 0) == 0)
+	    if (CVS_FNMATCH (*cpp++, name, 0) == 0)
 		return 1;
 	return 0;
     }
 }
 
-/* FIXME: This list of dirs to ignore stuff seems not to be used.  */
+/* FIXME: This list of dirs to ignore stuff seems not to be used.
+   Really?  send_dirent_proc and update_dirent_proc both call
+   ignore_directory and do_module calls ign_dir_add.  No doubt could
+   use some documentation/testsuite work.  */
 
 static char **dir_ign_list = NULL;
 static int dir_ign_max = 0;
 static int dir_ign_current = 0;
 
-/* add a directory to list of dirs to ignore */
-void ign_dir_add (name)
-     char *name;
+/* Add a directory to list of dirs to ignore.  */
+void
+ign_dir_add (name)
+    char *name;
 {
-  /* make sure we've got the space for the entry */
-  if (dir_ign_current <= dir_ign_max)
+    /* Make sure we've got the space for the entry.  */
+    if (dir_ign_current <= dir_ign_max)
     {
-      dir_ign_max += IGN_GROW;
-      dir_ign_list = (char **) xrealloc ((char *) dir_ign_list, (dir_ign_max+1) * sizeof(char*));
+	dir_ign_max += IGN_GROW;
+	dir_ign_list =
+	    (char **) xrealloc (dir_ign_list,
+				(dir_ign_max + 1) * sizeof (char *));
     }
 
-  dir_ign_list[dir_ign_current] = name;
-
-  dir_ign_current += 1 ;
+    dir_ign_list[dir_ign_current++] = xstrdup (name);
 }
 
 
-/* this function returns 1 (true) if the given directory name is part of
- * the list of directories to ignore
- */
+/* Return nonzero if NAME is part of the list of directories to ignore.  */
 
-int ignore_directory (name)
-     char *name;
+int
+ignore_directory (name)
+    char *name;
 {
-  int i;
+    int i;
 
-  if (!dir_ign_list)
-    return 0;
+    if (!dir_ign_list)
+	return 0;
 
-  i = dir_ign_current;
-  while (i--)
+    i = dir_ign_current;
+    while (i--)
     {
-      if (strncmp(name, dir_ign_list[i], strlen(dir_ign_list[i])) == 0)
-	return 1;
+	if (strncmp (name, dir_ign_list[i], strlen (dir_ign_list[i])) == 0)
+	    return 1;
     }
 
-  return 0;
+    return 0;
 }
 
 /*
@@ -348,6 +379,8 @@ ignore_files (ilist, entries, update_dir, proc)
     struct stat sb;
     char *file;
     char *xdir;
+    List *files;
+    Node *p;
 
     /* Set SUBDIRS if we have subdirectory information in ENTRIES.  */
     if (entries == NULL)
@@ -368,12 +401,18 @@ ignore_files (ilist, entries, update_dir, proc)
 
     dirp = CVS_OPENDIR (".");
     if (dirp == NULL)
+    {
+	error (0, errno, "cannot open current directory");
 	return;
+    }
 
     ign_add_file (CVSDOTIGNORE, 1);
     wrap_add_file (CVSDOTWRAPPER, 1);
 
-    while ((dp = readdir (dirp)) != NULL)
+    /* Make a list for the files.  */
+    files = getlist ();
+
+    while (errno = 0, (dp = CVS_READDIR (dirp)) != NULL)
     {
 	file = dp->d_name;
 	if (strcmp (file, ".") == 0 || strcmp (file, "..") == 0)
@@ -418,32 +457,53 @@ ignore_files (ilist, entries, update_dir, proc)
 
 	    if (
 #ifdef DT_DIR
-		dp->d_type == DT_DIR || dp->d_type == DT_UNKNOWN &&
+		dp->d_type == DT_DIR
+		|| (dp->d_type == DT_UNKNOWN && S_ISDIR (sb.st_mode))
+#else
+		S_ISDIR (sb.st_mode)
 #endif
-		S_ISDIR(sb.st_mode))
+		)
 	    {
 		if (! subdirs)
 		{
-		    char temp[PATH_MAX];
+		    char *temp;
 
+		    temp = xmalloc (strlen (file) + sizeof (CVSADM) + 10);
 		    (void) sprintf (temp, "%s/%s", file, CVSADM);
 		    if (isdir (temp))
+		    {
+			free (temp);
 			continue;
+		    }
+		    free (temp);
 		}
 	    }
 #ifdef S_ISLNK
 	    else if (
 #ifdef DT_DIR
-		dp->d_type == DT_LNK || dp->d_type == DT_UNKNOWN && 
+		     dp->d_type == DT_LNK
+		     || (dp->d_type == DT_UNKNOWN && S_ISLNK(sb.st_mode))
+#else
+		     S_ISLNK (sb.st_mode)
 #endif
-		S_ISLNK(sb.st_mode))
+		     )
 	    {
 		continue;
 	    }
 #endif
-    	}
+	}
 
-	(*proc) (file, xdir);
+	p = getnode ();
+	p->type = FILES;
+	p->key = xstrdup (file);
+	(void) addnode (files, p);
     }
-    (void) closedir (dirp);
+    if (errno != 0)
+	error (0, errno, "error reading current directory");
+    (void) CVS_CLOSEDIR (dirp);
+
+    sortlist (files, fsortcmp);
+    for (p = files->list->next; p != files->list; p = p->next)
+	(*proc) (p->key, xdir);
+    dellist (&files);
 }
