@@ -6,6 +6,8 @@
  * specified in the README file that comes with the CVS source distribution.
  */
 
+#include <assert.h>
+
 #include "cvs.h"
 #include "getline.h"
 
@@ -27,6 +29,15 @@ static char *editinfo_editor;
 static char *verifymsg_script;
 static Ctype type;
 
+/* 
+ * Should the logmsg be re-read during the do_verify phase?
+ * RereadLogAfterVerify=no|stat|yes
+ * LOGMSG_REREAD_NEVER  - never re-read the logmsg
+ * LOGMSG_REREAD_STAT   - re-read the logmsg only if it has changed
+ * LOGMSG_REREAD_ALWAYS - always re-read the logmsg
+ */
+int RereadLogAfterVerify = LOGMSG_REREAD_ALWAYS;
+
 /*
  * Puts a standard header on the output which is either being prepared for an
  * editor session, or being sent to a logfile program.  The modified, added,
@@ -166,7 +177,8 @@ fmt_proc (p, closure)
  * stripped and the log message is stored in the "message" argument.
  * 
  * If REPOSITORY is non-NULL, process rcsinfo for that repository; if it
- * is NULL, use the CVSADM_TEMPLATE file instead.
+ * is NULL, use the CVSADM_TEMPLATE file instead.  REPOSITORY should be
+ * NULL when running in client mode.
  */
 void
 do_editor (dir, messagep, repository, changes)
@@ -182,7 +194,9 @@ do_editor (dir, messagep, repository, changes)
     char *fname;
     struct stat pre_stbuf, post_stbuf;
     int retcode = 0;
-    char *p;
+
+    assert (current_parsed_root->isremote && !repository
+	    || !current_parsed_root->isremote && repository);
 
     if (noexec || reuse_log_message)
 	return;
@@ -191,8 +205,10 @@ do_editor (dir, messagep, repository, changes)
     if (strcmp (Editor, "") == 0 && !editinfo_editor)
 	error(1, 0, "no editor defined, must use -e or -m");
 
-
     /* Create a temporary file */
+    /* FIXME - It's possible we should be relying on cvs_temp_file to open
+     * the file here - we get race conditions otherwise.
+     */
     fname = cvs_temp_name ();
   again:
     if ((fp = CVS_FOPEN (fname, "w+")) == NULL)
@@ -206,8 +222,6 @@ do_editor (dir, messagep, repository, changes)
 	    (*messagep)[strlen (*messagep) - 1] != '\n')
 	    (void) fprintf (fp, "\n");
     }
-    else
-	(void) fprintf (fp, "\n");
 
     if (repository != NULL)
 	/* tack templates on if necessary */
@@ -216,7 +230,6 @@ do_editor (dir, messagep, repository, changes)
     {
 	FILE *tfp;
 	char buf[1024];
-	char *p;
 	size_t n;
 	size_t nwrite;
 
@@ -231,9 +244,9 @@ do_editor (dir, messagep, repository, changes)
 	{
 	    while (!feof (tfp))
 	    {
+		char *p = buf;
 		n = fread (buf, 1, sizeof buf, tfp);
 		nwrite = n;
-		p = buf;
 		while (nwrite > 0)
 		{
 		    n = fwrite (p, 1, nwrite, fp);
@@ -274,7 +287,7 @@ do_editor (dir, messagep, repository, changes)
 	free (editinfo_editor);
     editinfo_editor = (char *) NULL;
 #ifdef CLIENT_SUPPORT
-    if (client_active)
+    if (current_parsed_root->isremote)
 	; /* nothing, leave editinfo_editor NULL */
     else
 #endif
@@ -307,7 +320,7 @@ do_editor (dir, messagep, repository, changes)
 	/* On NT, we might read less than st_size bytes, but we won't
 	   read more.  So this works.  */
 	*messagep = (char *) xmalloc (post_stbuf.st_size + 1);
- 	*messagep[0] = '\0';
+ 	(*messagep)[0] = '\0';
     }
 
     line = NULL;
@@ -315,7 +328,8 @@ do_editor (dir, messagep, repository, changes)
 
     if (*messagep)
     {
-	p = *messagep;
+	size_t message_len = post_stbuf.st_size + 1;
+	size_t offset = 0;
 	while (1)
 	{
 	    line_length = getline (&line, &line_chars_allocated, fp);
@@ -327,8 +341,11 @@ do_editor (dir, messagep, repository, changes)
 	    }
 	    if (strncmp (line, CVSEDITPREFIX, CVSEDITPREFIXLEN) == 0)
 		continue;
-	    (void) strcpy (p, line);
-	    p += line_length;
+	    if (offset + line_length >= message_len)
+		expand_string (messagep, &message_len,
+				offset + line_length + 1);
+	    (void) strcpy (*messagep + offset, line);
+	    offset += line_length;
 	}
     }
     if (fclose (fp) < 0)
@@ -336,8 +353,14 @@ do_editor (dir, messagep, repository, changes)
 
     if (pre_stbuf.st_mtime == post_stbuf.st_mtime ||
 	*messagep == NULL ||
+	(*messagep)[0] == '\0' ||
 	strcmp (*messagep, "\n") == 0)
     {
+	if (*messagep)
+	{
+	    free (*messagep);
+	    *messagep = NULL;
+	}
 	for (;;)
 	{
 	    (void) printf ("\nLog message unchanged or not specified\n");
@@ -385,16 +408,18 @@ do_editor (dir, messagep, repository, changes)
    independant of the running of an editor for getting a message.
  */
 void
-do_verify (message, repository)
-    char *message;
+do_verify (messagep, repository)
+    char **messagep;
     char *repository;
 {
     FILE *fp;
     char *fname;
     int retcode = 0;
 
+    struct stat pre_stbuf, post_stbuf;
+
 #ifdef CLIENT_SUPPORT
-    if (client_active)
+    if (current_parsed_root->isremote)
 	/* The verification will happen on the server.  */
 	return;
 #endif
@@ -406,60 +431,133 @@ do_verify (message, repository)
 
     /* If there's no message, then we have nothing to verify.  Can this
        case happen?  And if so why would we print a message?  */
-    if (message == NULL)
+    if (*messagep == NULL)
     {
 	cvs_output ("No message to verify\n", 0);
 	return;
     }
 
-    /* Get a temp filename, open a temporary file, write the message to the 
+    /* open a temporary file, write the message to the 
        temp file, and close the file.  */
 
-    fname = cvs_temp_name ();
-
-    fp = fopen (fname, "w");
-    if (fp == NULL)
+    if ((fp = cvs_temp_file (&fname)) == NULL)
 	error (1, errno, "cannot create temporary file %s", fname);
-    else
+
+    fprintf (fp, "%s", *messagep);
+    if ((*messagep)[0] == '\0' ||
+	(*messagep)[strlen (*messagep) - 1] != '\n')
+	(void) fprintf (fp, "%s", "\n");
+    if (fclose (fp) == EOF)
+	error (1, errno, "%s", fname);
+
+    if (RereadLogAfterVerify == LOGMSG_REREAD_STAT)
     {
-	fprintf (fp, "%s", message);
-	if ((message)[0] == '\0' ||
-	    (message)[strlen (message) - 1] != '\n')
-	    (void) fprintf (fp, "%s", "\n");
-	if (fclose (fp) == EOF)
-	    error (1, errno, "%s", fname);
+	/* Remember the status of the temp file for later */
+	if ( CVS_STAT (fname, &pre_stbuf) != 0 )
+	    error (1, errno, "cannot stat temp file %s", fname);
 
-	/* Get the name of the verification script to run  */
+	/*
+	 * See if we need to sleep before running the verification
+	 * script to avoid time-stamp races.
+	 */
+	sleep_past (pre_stbuf.st_mtime);
+    }
 
-	if (repository != NULL)
-	    (void) Parse_Info (CVSROOTADM_VERIFYMSG, repository, 
-			       verifymsg_proc, 0);
+    /* Get the name of the verification script to run  */
 
-	/* Run the verification script  */
+    if (repository != NULL)
+	(void) Parse_Info (CVSROOTADM_VERIFYMSG, repository, 
+			   verifymsg_proc, 0);
 
-	if (verifymsg_script)
+    /* Run the verification script  */
+
+    if (verifymsg_script)
+    {
+	run_setup (verifymsg_script);
+	run_arg (fname);
+	if ((retcode = run_exec (RUN_TTY, RUN_TTY, RUN_TTY,
+				 RUN_NORMAL | RUN_SIGIGNORE)) != 0)
 	{
-	    run_setup (verifymsg_script);
-	    run_arg (fname);
-	    if ((retcode = run_exec (RUN_TTY, RUN_TTY, RUN_TTY,
-				     RUN_NORMAL | RUN_SIGIGNORE)) != 0)
-	    {
-		/* Since following error() exits, delete the temp file
-		   now.  */
-		if (unlink_file (fname) < 0)
-		    error (0, errno, "cannot remove %s", fname);
+	    /* Since following error() exits, delete the temp file now.  */
+	    if (unlink_file (fname) < 0)
+		error (0, errno, "cannot remove %s", fname);
 
-		error (1, retcode == -1 ? errno : 0, 
-		       "Message verification failed");
-	    }
+	    error (1, retcode == -1 ? errno : 0, 
+		   "Message verification failed");
+	}
+    }
+
+    /* Get the mod time and size of the possibly new log message
+     * in always and stat modes.
+     */
+    if (RereadLogAfterVerify == LOGMSG_REREAD_ALWAYS ||
+	RereadLogAfterVerify == LOGMSG_REREAD_STAT)
+    {
+	if ( CVS_STAT (fname, &post_stbuf) != 0 )
+	    error (1, errno, "cannot find size of temp file %s", fname);
+    }
+
+    /* And reread the log message in `always' mode or in `stat' mode when it's
+     * changed
+     */
+    if (RereadLogAfterVerify == LOGMSG_REREAD_ALWAYS ||
+	(RereadLogAfterVerify == LOGMSG_REREAD_STAT &&
+	    (pre_stbuf.st_mtime != post_stbuf.st_mtime ||
+	     pre_stbuf.st_size != post_stbuf.st_size)))
+    {
+	/* put the entire message back into the *messagep variable */
+	if ( (fp = open_file (fname, "r")) == NULL )
+	    error (1, errno, "cannot open temporary file %s", fname);
+
+	if (*messagep) free (*messagep);
+
+	if (post_stbuf.st_size == 0)
+	    *messagep = NULL;
+	else
+	{
+	    /* On NT, we might read less than st_size bytes,
+	       but we won't read more.  So this works.  */
+	    *messagep = (char *) xmalloc (post_stbuf.st_size + 1);
+	    *messagep[0] = '\0';
 	}
 
-	/* Delete the temp file  */
+	if (*messagep)
+	{
+	    char *line = NULL;
+	    int line_length;
+	    size_t line_chars_allocated = 0;
+	    char *p = *messagep;
 
-	if (unlink_file (fname) < 0)
-	    error (0, errno, "cannot remove %s", fname);
-	free (fname);
+	    while (1)
+	    {
+		line_length = getline (&line,
+				       &line_chars_allocated,
+				       fp);
+		if (line_length == -1)
+		{
+		    if (ferror (fp))
+			/* Fail in this case because otherwise we will have no
+			 * log message
+			 */
+			error (1, errno, "cannot read %s", fname);
+		    break;
+		}
+		if (strncmp (line, CVSEDITPREFIX, CVSEDITPREFIXLEN) == 0)
+		    continue;
+		(void) strcpy (p, line);
+		p += line_length;
+	    }
+	    if (line) free (line);
+	}
+	if (fclose (fp) < 0)
+	    error (0, errno, "warning: cannot close %s", fname);
     }
+
+    /* Delete the temp file  */
+
+    if (unlink_file (fname) < 0)
+	error (0, errno, "cannot remove %s", fname);
+    free (fname);
 }
 
 /*
@@ -759,7 +857,7 @@ logfile_write (repository, filter, message, logfp, changes)
 	}
 
 	len = fmt_end - fmt_begin;
-	str_list_format = xmalloc (sizeof (char) * (len + 1));
+	str_list_format = xmalloc (len + 1);
 	strncpy (str_list_format, fmt_begin, len);
 	str_list_format[len] = '\0';
 
@@ -790,16 +888,16 @@ logfile_write (repository, filter, message, logfp, changes)
 
 	srepos = Short_Repository (repository);
 
-	prog = xmalloc ((fmt_percent - filter) + strlen (srepos)
-			+ strlen (str_list) + strlen (fmt_continue)
+	prog = cp = xmalloc ((fmt_percent - filter) + 2 * strlen (srepos)
+			+ 2 * strlen (str_list) + strlen (fmt_continue)
 			+ 10);
-	(void) strncpy (prog, filter, fmt_percent - filter);
-	prog[fmt_percent - filter] = '\0';
-	(void) strcat (prog, "'");
-	(void) strcat (prog, srepos);
-	(void) strcat (prog, str_list);
-	(void) strcat (prog, "'");
-	(void) strcat (prog, fmt_continue);
+	(void) memcpy (cp, filter, fmt_percent - filter);
+	cp += fmt_percent - filter;
+	*cp++ = '"';
+	cp = shell_escape (cp, srepos);
+	cp = shell_escape (cp, str_list);
+	*cp++ = '"';
+	(void) strcpy (cp, fmt_continue);
 	    
 	/* To be nice, free up some memory. */
 
