@@ -206,10 +206,12 @@ static const char *const checkoutlist_contents[] = {
 static const char *const cvswrappers_contents[] = {
     "# This file affects handling of files based on their names.\n",
     "#\n",
+#if 0    /* see comments in wrap_add in wrapper.c */
     "# The -t/-f options allow one to treat directories of files\n",
     "# as a single file, or to transform a file in other ways on\n",
     "# its way in and out of CVS.\n",
     "#\n",
+#endif
     "# The -m option specifies whether CVS attempts to merge files.\n",
     "#\n",
     "# The -k option specifies keyword expansion (e.g. -kb for binary).\n",
@@ -242,7 +244,7 @@ static const char *const notify_contents[] = {
     "# \"ALL\" or \"DEFAULT\" can be used in place of the regular expression.\n",
     "#\n",
     "# For example:\n",
-    "#ALL mail %s -s \"CVS notification\"\n",
+    "#ALL mail -s \"CVS notification\" %s\n",
     NULL
 };
 
@@ -297,6 +299,14 @@ static const char *const config_contents[] = {
     "# Set `LogHistory' to `all' or `TOFEWGCMAR' to log all transactions to the\n",
     "# history file, or a subset as needed (ie `TMAR' logs all write operations)\n",
     "#LogHistory=TOFEWGCMAR\n",
+    "\n",
+    "# Set `RereadLogAfterVerify' to `always' (the default) to allow the verifymsg\n",
+    "# script to change the log message.  Set it to `stat' to force CVS to verify",
+    "# that the file has changed before reading it (this can take up to an extra\n",
+    "# second per directory being committed, so it is not recommended for large\n",
+    "# repositories.  Set it to `never' (the previous CVS behavior) to prevent\n",
+    "# verifymsg scripts from changing the log message.\n",
+    "#RereadLogAfterVerify=always\n",
     NULL
 };
 
