@@ -147,7 +147,8 @@ call_diff_write_output (text, len)
     const char *text;
     size_t len;
 {
-    cvs_output (text, len);
+    if (len > 0)
+	cvs_output (text, len);
 }
 
 /* Call back function for the diff library to flush the output file.
@@ -529,9 +530,11 @@ RCS file: ", 0);
    message on stderr.  */
 
 int
-diff_exec (file1, file2, options, out)
+diff_exec (file1, file2, label1, label2, options, out)
     char *file1;
     char *file2;
+    char *label1;
+    char *label2;
     char *options;
     char *out;
 {
@@ -574,6 +577,10 @@ diff_exec (file1, file2, options, out)
     /* The first word in this string is used only for error reporting. */
     sprintf (args, "diff %s", options);
     call_diff_setup (args);
+    if (label1)
+	call_diff_arg (label1);
+    if (label2)
+	call_diff_arg (label2);
     call_diff_arg (file1);
     call_diff_arg (file2);
     free (args);
