@@ -157,10 +157,10 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 #ifdef CLIENT_SUPPORT
 	if (!just_subdirs
 	    && CVSroot_cmdline == NULL
-	    && client_active)
+	    && current_parsed_root->isremote)
 	{
 	    char *root = Name_Root (NULL, update_dir);
-	    if (root && strcmp (root, current_root) != 0)
+	    if (root && strcmp (root, current_parsed_root->original) != 0)
 		/* We're skipping this directory because it is for
 		   a different root.  Therefore, we just want to
 		   do the subdirectories only.  Processing files would
@@ -173,6 +173,7 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 		   seems to be handled somewhere (else) but why should
 		   it be a separate case?  Needs investigation...  */
 		just_subdirs = 1;
+	    free (root);
 	}
 #endif
 
@@ -203,7 +204,7 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 		       program_name);
 	    }
 #ifdef CLIENT_SUPPORT
-	    else if (client_active && server_started)
+	    else if (current_parsed_root->isremote && server_started)
 	    {
 		/* In the the case "cvs update foo bar baz", a call to
 		   send_file_names in update.c will have sent the
@@ -246,7 +247,10 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 	   directories. */
 
 	if (!wrap_name_has (argv[i], WRAP_TOCVS) && isdir (argv[i]))
+	{
+	    strip_trailing_slashes (argv[i]);
 	    addlist (&dirlist, argv[i]);
+	}
 	else
 	{
 	    /* otherwise, split argument into directory and component names. */
@@ -289,7 +293,7 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 	    {
 		if ((which & W_LOCAL) && isdir (CVSADM)
 #ifdef CLIENT_SUPPORT
-		    && !client_active
+		    && !current_parsed_root->isremote
 #endif
 		    )
 		{
@@ -362,8 +366,8 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 	/* FIXME (njc): in the multiroot case, we don't want to send
 	   argument commands for those top-level directories which do
 	   not contain any subdirectories which have files checked out
-	   from current_root.  If we do, and two repositories have a
-	   module with the same name, nasty things could happen.
+	   from current_parsed_root->original.  If we do, and two repositories
+	   have a module with the same name, nasty things could happen.
 
 	   This is hard.  Perhaps we should send the Argument commands
 	   later in this procedure, after we've had a chance to notice
@@ -439,7 +443,7 @@ start_recursion (fileproc, filesdoneproc, direntproc, dirleaveproc, callerdat,
 	   "Directory xxx" command, which forces the server to descend
 	   and serve the files there.  client.c (send_file_names) has
 	   also been modified to send only those arguments which are
-	   appropriate to current_root.
+	   appropriate to current_parsed_root->original.
 
 	*/
 		
@@ -590,7 +594,7 @@ do_recursion (frame)
 		/* Add it to our list. */
 
 		Node *n = getnode ();
-		n->type = UNKNOWN;
+		n->type = NT_UNKNOWN;
 		n->key = xstrdup (this_root);
 
 		if (addnode (root_directories, n))
@@ -598,8 +602,9 @@ do_recursion (frame)
 	
 	    }
 	
-	    process_this_directory = (strcmp (current_root, this_root) == 0);
-	
+	    process_this_directory =
+		    (strcmp (current_parsed_root->original, this_root) == 0);
+
 	    free (this_root);
 	}
     }
@@ -709,7 +714,7 @@ do_recursion (frame)
 	   place (server_notify).  For local, we can't do them here--we don't
 	   have writelocks in place, and there is no way to get writelocks
 	   here.  */
-	if (client_active)
+	if (current_parsed_root->isremote)
 	    notify_check (repository, update_dir);
 #endif /* CLIENT_SUPPORT */
 
@@ -1015,7 +1020,7 @@ but CVS uses %s for its own purposes; skipping %s directory",
 		/* Add it to our list. */
 
 		Node *n = getnode ();
-		n->type = UNKNOWN;
+		n->type = NT_UNKNOWN;
 		n->key = xstrdup (this_root);
 
 		if (addnode (root_directories, n))
@@ -1023,7 +1028,8 @@ but CVS uses %s for its own purposes; skipping %s directory",
 
 	    }
 
-	    process_this_directory = (strcmp (current_root, this_root) == 0);
+	    process_this_directory = (strcmp (current_parsed_root->original, this_root) == 0);
+
 	    free (this_root);
 	}
     }
@@ -1132,6 +1138,7 @@ addfile (listp, dir, file)
     char *file;
 {
     Node *n;
+    List *fl;
 
     /* add this dir. */
     addlist (listp, dir);
@@ -1144,7 +1151,9 @@ addfile (listp, dir, file)
     }
 
     n->type = DIRS;
-    addlist ((List **) &n->data, file);
+    fl = (List *) n->data;
+    addlist (&fl, file);
+    n->data = (char *) fl;
     return;
 }
 
@@ -1204,6 +1213,7 @@ unroll_files_proc (p, closure)
     }
 
     dirlist = save_dirlist;
-    filelist = NULL;
+    if (filelist)
+	dellist (&filelist);
     return(err);
 }
