@@ -3,7 +3,7 @@
  * Copyright (c) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
- * specified in the README file that comes with the CVS 1.4 kit.
+ * specified in the README file that comes with the CVS source distribution.
  * 
  * Remove a File
  * 
@@ -17,6 +17,10 @@
 
 #include "cvs.h"
 
+#ifdef CLIENT_SUPPORT
+static int remove_force_fileproc PROTO ((void *callerdat,
+					 struct file_info *finfo));
+#endif
 static int remove_fileproc PROTO ((void *callerdat, struct file_info *finfo));
 static Dtype remove_dirproc PROTO ((void *callerdat, char *dir,
 				    char *repos, char *update_dir,
@@ -33,6 +37,7 @@ static const char *const remove_usage[] =
     "\t-f\tDelete the file before removing it.\n",
     "\t-l\tProcess this directory only (not recursive).\n",
     "\t-R\tProcess directories recursively.\n",
+    "(Specify the --help global option for a list of other help options)\n",
     NULL
 };
 
@@ -46,8 +51,8 @@ cvsremove (argc, argv)
     if (argc == -1)
 	usage (remove_usage);
 
-    optind = 1;
-    while ((c = getopt (argc, argv, "flR")) != -1)
+    optind = 0;
+    while ((c = getopt (argc, argv, "+flR")) != -1)
     {
 	switch (c)
 	{
@@ -72,7 +77,7 @@ cvsremove (argc, argv)
     wrap_setup ();
 
 #ifdef CLIENT_SUPPORT
-    if (client_active) {
+    if (current_parsed_root->isremote) {
 	/* Call expand_wild so that the local removal of files will
            work.  It's ok to do it always because we have to send the
            file names expanded anyway.  */
@@ -82,15 +87,10 @@ cvsremove (argc, argv)
 	{
 	    if (!noexec)
 	    {
-		int i;
-
-		for (i = 0; i < argc; i++)
-		{
-		    if ( CVS_UNLINK (argv[i]) < 0 && ! existence_error (errno))
-		    {
-			error (0, errno, "unable to remove %s", argv[i]);
-		    }
-		}
+		start_recursion (remove_force_fileproc, (FILESDONEPROC) NULL,
+				 (DIRENTPROC) NULL, (DIRLEAVEPROC) NULL,
+				 (void *) NULL, argc, argv, local, W_LOCAL,
+				 0, 0, (char *) NULL, 0);
 	    }
 	    /* else FIXME should probably act as if the file doesn't exist
 	       in doing the following checks.  */
@@ -100,8 +100,10 @@ cvsremove (argc, argv)
 	ign_setup ();
 	if (local)
 	    send_arg("-l");
+	/* FIXME: Can't we set SEND_NO_CONTENTS here?  Needs investigation.  */
+	send_files (argc, argv, local, 0, 0);
 	send_file_names (argc, argv, 0);
-	send_files (argc, argv, local, 0);
+	free_names (&argc, argv);
 	send_to_server ("remove\012", 0);
         return get_responses_and_close ();
     }
@@ -113,7 +115,7 @@ cvsremove (argc, argv)
 			   argc, argv,
                            local, W_LOCAL, 0, 1, (char *) NULL, 1);
 
-    if (removed_files)
+    if (removed_files && !really_quiet)
 	error (0, 0, "use '%s commit' to remove %s permanently", program_name,
 	       (removed_files == 1) ? "this file" : "these files");
 
@@ -127,6 +129,26 @@ cvsremove (argc, argv)
     return (err);
 }
 
+#ifdef CLIENT_SUPPORT
+
+/*
+ * This is called via start_recursion if we are running as the client
+ * and the -f option was used.  We just physically remove the file.
+ */
+
+/*ARGSUSED*/
+static int
+remove_force_fileproc (callerdat, finfo)
+     void *callerdat;
+     struct file_info *finfo;
+{
+    if (CVS_UNLINK (finfo->file) < 0 && ! existence_error (errno))
+	error (0, errno, "unable to remove %s", finfo->fullname);
+    return 0;
+}
+
+#endif
+
 /*
  * remove the file, only if it has already been physically removed
  */
@@ -136,7 +158,6 @@ remove_fileproc (callerdat, finfo)
     void *callerdat;
     struct file_info *finfo;
 {
-    char fname[PATH_MAX];
     Vers_TS *vers;
 
     if (force)
@@ -168,13 +189,21 @@ remove_fileproc (callerdat, finfo)
     }
     else if (vers->vn_user[0] == '0' && vers->vn_user[1] == '\0')
     {
+	char *fname;
+
 	/*
 	 * It's a file that has been added, but not commited yet. So,
 	 * remove the ,t file for it and scratch it from the
 	 * entries file.  */
 	Scratch_Entry (finfo->entries, finfo->file);
+	fname = xmalloc (strlen (finfo->file)
+			 + sizeof (CVSADM)
+			 + sizeof (CVSEXT_LOG)
+			 + 10);
 	(void) sprintf (fname, "%s/%s%s", CVSADM, finfo->file, CVSEXT_LOG);
-	(void) unlink_file (fname);
+	if (unlink_file (fname) < 0
+	    && !existence_error (errno))
+	    error (0, errno, "cannot remove %s", CVSEXT_LOG);
 	if (!quiet)
 	    error (0, 0, "removed `%s'", finfo->fullname);
 
@@ -182,6 +211,7 @@ remove_fileproc (callerdat, finfo)
 	if (server_active)
 	    server_checked_in (finfo->file, finfo->update_dir, finfo->repository);
 #endif
+	free (fname);
     }
     else if (vers->vn_user[0] == '-')
     {
@@ -189,9 +219,36 @@ remove_fileproc (callerdat, finfo)
 	    error (0, 0, "file `%s' already scheduled for removal",
 		   finfo->fullname);
     }
+    else if (vers->tag != NULL && isdigit ((unsigned char) *vers->tag))
+    {
+	/* Commit will just give an error, and so there seems to be
+	   little reason to allow the remove.  I mean, conflicts that
+	   arise out of parallel development are one thing, but conflicts
+	   that arise from sticky tags are quite another.
+
+	   I would have thought that non-branch sticky tags should be the
+	   same but at least now, removing a file with a non-branch sticky
+	   tag means to delete the tag from the file.  I'm not sure that
+	   is a good behavior, but until it is changed, we need to allow
+	   it.  */
+	error (0, 0, "\
+cannot remove file `%s' which has a numeric sticky tag of `%s'",
+	       finfo->fullname, vers->tag);
+    }
+    else if (vers->date != NULL)
+    {
+	/* Commit will just give an error, and so there seems to be
+	   little reason to allow the remove.  */
+	error (0, 0, "\
+cannot remove file `%s' which has a sticky date of `%s'",
+	       finfo->fullname, vers->date);
+    }
     else
     {
+	char *fname;
+
 	/* Re-register it with a negative version number.  */
+	fname = xmalloc (strlen (vers->vn_user) + 5);
 	(void) strcpy (fname, "-");
 	(void) strcat (fname, vers->vn_user);
 	Register (finfo->entries, finfo->file, fname, vers->ts_rcs, vers->options,
@@ -204,6 +261,7 @@ remove_fileproc (callerdat, finfo)
 	if (server_active)
 	    server_checked_in (finfo->file, finfo->update_dir, finfo->repository);
 #endif
+	free (fname);
     }
 
     freevers_ts (&vers);
