@@ -11,6 +11,15 @@
 #include "cvs.h"
 #include "getline.h"
 
+#ifdef HAVE_NANOSLEEP
+# include "xtime.h"
+#else /* HAVE_NANOSLEEP */
+# if !defined HAVE_USLEEP && defined HAVE_SELECT
+    /* use select as a workaround */
+#   include "xselect.h"
+# endif /* !defined HAVE_USLEEP && defined HAVE_SELECT */
+#endif /* !HAVE_NANOSLEEP */
+
 extern char *getlogin ();
 
 /*
@@ -111,6 +120,19 @@ expand_string (strptr, n, newsize)
     }
 }
 
+/* *STR is a pointer to a malloc'd string.  *LENP is its allocated
+   length.  Add SRC to the end of it, reallocating if necessary.  */
+void
+xrealloc_and_strcat (str, lenp, src)
+    char **str;
+    size_t *lenp;
+    const char *src;
+{
+
+    expand_string (str, lenp, strlen (*str) + strlen (src) + 1);
+    strcat (*str, src);
+}
+
 /*
  * Duplicate a string, calling xmalloc to allocate some dynamic space
  */
@@ -171,7 +193,8 @@ pathname_levels (path)
 	    if (-level > max_level)
 		max_level = -level;
 	}
-	else if (p[0] == '.' && (p[1] == '\0' || p[1] == '/'))
+	else if (p[0] == '\0' || p[0] == '/' ||
+		 (p[0] == '.' && (p[1] == '\0' || p[1] == '/')))
 	    ;
 	else
 	    ++level;
@@ -218,9 +241,7 @@ line2argv (pargc, argv, line, sepchars)
     int argv_allocated;
 
     /* Small for testing.  */
-    /* argv_allocated must be at least 3 because at some places
-       (e.g. checkout_proc) cvs alters argv[2].  */
-    argv_allocated = 4;
+    argv_allocated = 1;
     *argv = (char **) xmalloc (argv_allocated * sizeof (**argv));
 
     *pargc = 0;
@@ -586,7 +607,9 @@ file_has_markers (finfo)
 	error (1, errno, "cannot open %s", finfo->fullname);
     while (getline (&line, &line_allocated, fp) > 0)
     {
-	if (strncmp (line, RCS_MERGE_PAT, sizeof RCS_MERGE_PAT - 1) == 0)
+	if (strncmp (line, RCS_MERGE_PAT_1, sizeof RCS_MERGE_PAT_1 - 1) == 0 ||
+	    strncmp (line, RCS_MERGE_PAT_2, sizeof RCS_MERGE_PAT_2 - 1) == 0 ||
+	    strncmp (line, RCS_MERGE_PAT_3, sizeof RCS_MERGE_PAT_3 - 1) == 0)
 	{
 	    result = 1;
 	    goto out;
@@ -647,9 +670,9 @@ get_file (name, fullname, mode, buf, bufsize, len)
 	e = open_file (name, mode);
     }
 
-    if (*bufsize < filesize)
+    if (*buf == NULL || *bufsize <= filesize)
     {
-	*bufsize = filesize;
+	*bufsize = filesize + 1;
 	*buf = xrealloc (*buf, *bufsize);
     }
 
@@ -691,12 +714,9 @@ get_file (name, fullname, mode, buf, bufsize, len)
     *len = nread;
 
     /* Force *BUF to be large enough to hold a null terminator. */
-    if (*buf != NULL)
-    {
-	if (nread == *bufsize)
-	    expand_string (buf, bufsize, *bufsize + 1);
-	(*buf)[nread] = '\0';
-    }
+    if (nread == *bufsize)
+	expand_string (buf, bufsize, *bufsize + 1);
+    (*buf)[nread] = '\0';
 }
 
 
@@ -722,9 +742,6 @@ resolve_symlink (filename)
 	   But that would require editing each filesubr.c and so the
 	   expedient hack seems to be looking at HAVE_READLINK.  */
 	newname = xreadlink (*filename);
-#else
-	error (1, 0, "internal error: islink doesn't like readlink");
-#endif
 	
 	if (isabsolute (newname))
 	{
@@ -742,5 +759,135 @@ resolve_symlink (filename)
 	    free (*filename);
 	    *filename = fullnewname;
 	}
+#else
+	error (1, 0, "internal error: islink doesn't like readlink");
+#endif
+    }
+}
+
+/*
+ * Rename a file to an appropriate backup name based on BAKPREFIX.
+ * If suffix non-null, then ".<suffix>" is appended to the new name.
+ *
+ * Returns the new name, which caller may free() if desired.
+ */
+char *
+backup_file (filename, suffix)
+     const char *filename;
+     const char *suffix;
+{
+    char *backup_name;
+
+    if (suffix == NULL)
+    {
+        backup_name = xmalloc (sizeof (BAKPREFIX) + strlen (filename) + 1);
+        sprintf (backup_name, "%s%s", BAKPREFIX, filename);
+    }
+    else
+    {
+        backup_name = xmalloc (sizeof (BAKPREFIX)
+                               + strlen (filename)
+                               + strlen (suffix)
+                               + 2);  /* one for dot, one for trailing '\0' */
+        sprintf (backup_name, "%s%s.%s", BAKPREFIX, filename, suffix);
+    }
+
+    if (isfile (filename))
+        copy_file (filename, backup_name);
+
+    return backup_name;
+}
+
+/*
+ * Copy a string into a buffer escaping any shell metacharacters.  The
+ * buffer should be at least twice as long as the string.
+ *
+ * Returns a pointer to the terminating NUL byte in buffer.
+ */
+
+char *
+shell_escape(buf, str)
+    char *buf;
+    const char *str;
+{
+    static const char meta[] = "$`\\\"";
+    const char *p;
+
+    for (;;)
+    {
+	p = strpbrk(str, meta);
+	if (!p) p = str + strlen(str);
+	if (p > str)
+	{
+	    memcpy(buf, str, p - str);
+	    buf += p - str;
+	}
+	if (!*p) break;
+	*buf++ = '\\';
+	*buf++ = *p++;
+	str = p;
+    }
+    *buf = '\0';
+    return buf;
+}
+
+/*
+ * We can only travel forwards in time, not backwards.  :)
+ */
+void
+sleep_past (desttime)
+    time_t desttime;
+{
+    time_t t;
+    long s;
+    long us;
+
+    while (time (&t) <= desttime)
+    {
+#ifdef HAVE_GETTIMEOFDAY
+	struct timeval tv;
+	gettimeofday (&tv, NULL);
+	if (tv.tv_sec > desttime)
+	    break;
+	s = desttime - tv.tv_sec;
+	if (tv.tv_usec > 0)
+	    us = 1000000 - tv.tv_usec;
+	else
+	{
+	    s++;
+	    us = 0;
+	}
+#else
+	/* default to 20 ms increments */
+	s = desttime - t;
+	us = 20000;
+#endif
+
+#if defined(HAVE_NANOSLEEP)
+	{
+	    struct timespec ts;
+	    ts.tv_sec = s;
+	    ts.tv_nsec = us * 1000;
+	    (void)nanosleep (&ts, NULL);
+	}
+#elif defined(HAVE_USLEEP)
+	if (s > 0)
+	    (void)sleep (s);
+	else
+	    (void)usleep (us);
+#elif defined(HAVE_SELECT)
+	{
+	    /* use select instead of sleep since it is a fairly portable way of
+	     * sleeping for ms.
+	     */
+	    struct timeval tv;
+	    tv.tv_sec = s;
+	    tv.tv_usec = us;
+	    (void)select (0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &tv);
+	}
+#else
+	if (us > 0) s++;
+	(void)sleep(s);
+#endif
     }
 }
