@@ -87,22 +87,17 @@ Version_TS (finfo, options, tag, date, force_tag_match, set_time)
 	    vers_ts->vn_user = xstrdup (entdata->version);
 	    vers_ts->ts_rcs = xstrdup (entdata->timestamp);
 	    vers_ts->ts_conflict = xstrdup (entdata->conflict);
-	    if (!tag)
+	    if (!(tag || date) && !(sdtp && sdtp->aflag))
 	    {
-		if (!(sdtp && sdtp->aflag))
-		    vers_ts->tag = xstrdup (entdata->tag);
-	    }
-	    if (!date)
-	    {
-		if (!(sdtp && sdtp->aflag))
-		    vers_ts->date = xstrdup (entdata->date);
+		vers_ts->tag = xstrdup (entdata->tag);
+		vers_ts->date = xstrdup (entdata->date);
 	    }
 	    vers_ts->entdata = entdata;
 	}
 	/* Even if we don't have an "entries line" as such
 	   (vers_ts->entdata), we want to pick up options which could
 	   have been from a Kopt protocol request.  */
-	if (!options || (options && *options == '\0'))
+	if (!options || *options == '\0')
 	{
 	    if (!(sdtp && sdtp->aflag))
 		vers_ts->options = xstrdup (entdata->options);
@@ -126,6 +121,8 @@ Version_TS (finfo, options, tag, date, force_tag_match, set_time)
 	    char *rcsexpand = RCS_getexpand (finfo->rcs);
 	    if (rcsexpand != NULL)
 	    {
+		if (vers_ts->options != NULL)
+		    free (vers_ts->options);
 		vers_ts->options = xmalloc (strlen (rcsexpand) + 3);
 		strcpy (vers_ts->options, "-k");
 		strcat (vers_ts->options, rcsexpand);
@@ -325,6 +322,8 @@ time_stamp_server (file, vers_ts, entdata)
 	    cp = ctime (&sb.st_mtime);
 
 	cp[24] = 0;
+	/* Fix non-standard format.  */
+	if (cp[8] == '0') cp[8] = ' ';
 	(void) strcpy (vers_ts->ts_user, cp);
     }
 }
@@ -340,13 +339,22 @@ time_stamp (file)
 {
     struct stat sb;
     char *cp;
-    char *ts;
+    char *ts = NULL;
+    time_t mtime = 0L;
 
-    if (CVS_LSTAT (file, &sb) < 0)
+    if (!CVS_LSTAT (file, &sb))
     {
-	ts = NULL;
+	mtime = sb.st_mtime;
     }
-    else
+    /* If it's a symlink, return whichever is the newest mtime of
+       the link and its target, for safety.
+    */
+    if (!CVS_STAT (file, &sb))
+    {
+        if (mtime < sb.st_mtime)
+	    mtime = sb.st_mtime;
+    }
+    if (mtime)
     {
 	struct tm *tm_p;
         struct tm local_tm;
@@ -368,6 +376,8 @@ time_stamp (file)
 	    cp = ctime(&sb.st_mtime);
 
 	cp[24] = 0;
+	/* Fix non-standard format.  */
+	if (cp[8] == '0') cp[8] = ' ';
 	(void) strcpy (ts, cp);
     }
 
