@@ -17,6 +17,7 @@
    GNU General Public License for more details.  */
 
 #include "cvs.h"
+#include <assert.h>
 
 static int deep_remove_dir PROTO((const char *path));
 
@@ -405,42 +406,6 @@ rename_file (from_file, to_file)
 	error (1, errno, "cannot rename file %s to %s", from, to);
 }
 
-/*
- * link a file, if possible.
- */
-int
-link_file (from_file, to_file)
-    const char *from_file;
-    const char *to_file;
-{
-    char from[PATH_MAX], to[PATH_MAX];
-
-    /* Prefer local relative paths to files at expense of logical name
-       access to files. */
-
-    if (isabsolute(from_file))
-      strcpy(from, from_file);
-    else
-      sprintf(from, "./%s", from_file);
-
-    if (isabsolute(to_file))
-      strcpy(to, to_file);
-    else
-      sprintf(to, "./%s", to_file);
-
-    if (trace)
-#ifdef SERVER_SUPPORT
-	(void) fprintf (stderr, "%c-> link(%s,%s)\n",
-			(server_active) ? 'S' : ' ', from, to);
-#else
-	(void) fprintf (stderr, "-> link(%s,%s)\n", from, to);
-#endif
-    if (noexec)
-	return (0);
-
-    return (link (from, to));
-}
-
 /*
  * unlink a file, if possible.
  */
@@ -468,7 +433,7 @@ unlink_file (f_file)
     if (noexec)
 	return (0);
 
-    return (unlink (f));
+    return (vms_unlink (f));
 }
 
 /*
@@ -500,7 +465,7 @@ unlink_file_dir (f_file)
     if (noexec)
 	return (0);
 
-    if (unlink (f) != 0)
+    if (vms_unlink (f) != 0)
     {
 	/* under NEXTSTEP errno is set to return EPERM if
 	 * the file is a directory,or if the user is not
@@ -535,13 +500,13 @@ deep_remove_dir (path)
 
     if (rmdir (path) != 0 && (errno == ENOTEMPTY || errno == EEXIST)) 
     {
-	if ((dirp = opendir (path)) == NULL)
+	if ((dirp = CVS_OPENDIR (path)) == NULL)
 	    /* If unable to open the directory return
 	     * an error
 	     */
 	    return -1;
 
-	while ((dp = readdir (dirp)) != NULL)
+	while ((dp = CVS_READDIR (dirp)) != NULL)
 	{
 	    if (strcmp (dp->d_name, ".") == 0 ||
 			strcmp (dp->d_name, "..") == 0)
@@ -549,13 +514,13 @@ deep_remove_dir (path)
 
 	    sprintf (buf, "%s/%s", path, dp->d_name);
 
-	    if (unlink (buf) != 0 )
+	    if (vms_unlink (buf) != 0 )
 	    {
 		if (errno == EISDIR || errno == EPERM)
 		{
 		    if (deep_remove_dir (buf))
 		    {
-			closedir (dirp);
+			CVS_CLOSEDIR (dirp);
 			return -1;
 		    }
 		}
@@ -564,12 +529,12 @@ deep_remove_dir (path)
 		    /* buf isn't a directory, or there are
 		     * some sort of permision problems
 		     */
-		    closedir (dirp);
+		    CVS_CLOSEDIR (dirp);
 		    return -1;
 		}
 	    }
 	}
-	closedir (dirp);
+	CVS_CLOSEDIR (dirp);
 	return rmdir (path);
 	}
 
@@ -677,7 +642,7 @@ xcmp (file1_file, file2_file)
 	    if (read2 == (size_t)-1)
 		error (1, errno, "cannot read file %s for comparing", file2);
 
-	    /* assert (read1 == read2); */
+	    assert (read1 == read2);
 
 	    ret = memcmp(buf1, buf2, read1);
 	} while (ret == 0 && read1 == buf_size);
@@ -737,7 +702,7 @@ fncmp (const char *n1, const char *n2)
 {
     while (*n1 && *n2
            && (VMS_filename_classes[(unsigned char) *n1]
-	       == VMS_filename_classes[(unsigned char) *n2]))
+               == VMS_filename_classes[(unsigned char) *n2]))
         n1++, n2++;
     return (VMS_filename_classes[(unsigned char) *n1]
             - VMS_filename_classes[(unsigned char) *n2]);
@@ -769,19 +734,165 @@ fnfold (char *filename)
 }
 
 /* Generate a unique temporary filename.  Returns a pointer to a newly
-   malloc'd string containing the name.  Returns successfully or not at
-   all.  */
+ * malloc'd string containing the name.  Returns successfully or not at
+ * all.
+ *
+ *     THIS FUNCTION IS DEPRECATED!!!  USE cvs_temp_file INSTEAD!!!
+ *
+ * and yes, I know about the way the rcs commands use temp files.  I think
+ * they should be converted too but I don't have time to look into it right
+ * now.
+ */
 char *
 cvs_temp_name ()
 {
-    char value[L_tmpnam + 1];
-    char *retval;
+    char *fn;
+    FILE *fp;
 
-    /* FIXME: what is the VMS equivalent to TMPDIR?  */
-    retval = tmpnam (value);
-    if (retval == NULL)
-	error (1, errno, "cannot generate temporary filename");
-    return xstrdup (retval);
+    fp = cvs_temp_file (&fn);
+    if (fp == NULL)
+	error (1, errno, "Failed to create temporary file");
+    if (fclose (fp) == EOF)
+	error (0, errno, "Failed to close temporary file %s", fn);
+    return fn;
+}
+
+/* Generate a unique temporary filename and return an open file stream
+ * to the truncated file by that name
+ *
+ *  INPUTS
+ *	filename	where to place the pointer to the newly allocated file
+ *   			name string
+ *
+ *  OUTPUTS
+ *	filename	dereferenced, will point to the newly allocated file
+ *			name string.  This value is undefined if the function
+ *			returns an error.
+ *
+ *  RETURNS
+ *	An open file pointer to a read/write mode empty temporary file with the
+ *	unique file name or NULL on failure.
+ *
+ *  ERRORS
+ *	on error, errno will be set to some value either by CVS_FOPEN or
+ *	whatever system function is called to generate the temporary file name
+ */
+/* There are at least four functions for generating temporary
+ * filenames.  We use mkstemp (BSD 4.3) if possible, else tempnam (SVID 3),
+ * else mktemp (BSD 4.3), and as last resort tmpnam (POSIX).  Reason is that
+ * mkstemp, tempnam, and mktemp both allow to specify the directory in which
+ * the temporary file will be created.
+ *
+ * And the _correct_ way to use the deprecated functions probably involves
+ * opening file descriptors using O_EXCL & O_CREAT and even doing the annoying
+ * NFS locking thing, but until I hear of more problems, I'm not going to
+ * bother.
+ */
+FILE *cvs_temp_file (filename)
+    char **filename;
+{
+    char *fn;
+    FILE *fp;
+
+    /* FIXME - I'd like to be returning NULL here in noexec mode, but I think
+     * some of the rcs & diff functions which rely on a temp file run in
+     * noexec mode too.
+     */
+
+    assert (filename != NULL);
+
+#ifdef HAVE_MKSTEMP
+
+    {
+    int fd;
+
+    fn = xmalloc (strlen (Tmpdir) + 11);
+    sprintf (fn, "%s/%s", Tmpdir, "cvsXXXXXX" );
+    fd = mkstemp (fn);
+
+    /* a NULL return will be interpreted by callers as an error and
+     * errno should still be set
+     */
+    if (fd == -1) fp = NULL;
+    else if ((fp = CVS_FDOPEN (fd, "w+")) == NULL)
+    {
+	/* attempt to close and unlink the file since mkstemp returned sucessfully and
+	 * we believe it's been created and opened
+	 */
+ 	int save_errno = errno;
+	if (close (fd))
+	    error (0, errno, "Failed to close temporary file %s", fn);
+	if (CVS_UNLINK (fn))
+	    error (0, errno, "Failed to unlink temporary file %s", fn);
+	errno = save_errno;
+    }
+
+    if (fp == NULL) free (fn);
+    /* mkstemp is defined to open mode 0600 using glibc 2.0.7+ */
+    /* FIXME - configure can probably tell us which version of glibc we are
+     * linking to and not chmod for 2.0.7+
+     */
+    else chmod (fn, 0600);
+
+    }
+
+#elif HAVE_TEMPNAM
+
+    /* tempnam has been deprecated due to under-specification */
+
+    fn = tempnam (Tmpdir, "cvs");
+    if (fn == NULL) fp = NULL;
+    else if ((fp = CVS_FOPEN (fn, "w+")) == NULL) free (fn);
+    else chmod (fn, 0600);
+
+    /* tempnam returns a pointer to a newly malloc'd string, so there's
+     * no need for a xstrdup
+     */
+
+#elif HAVE_MKTEMP
+
+    /* mktemp has been deprecated due to the BSD 4.3 specification specifying
+     * that XXXXXX will be replaced by a PID and a letter, creating only 26
+     * possibilities, a security risk, and a race condition.
+     */
+
+    {
+    char *ifn;
+
+    ifn = xmalloc (strlen (Tmpdir) + 11);
+    sprintf (ifn, "%s/%s", Tmpdir, "cvsXXXXXX" );
+    fn = mktemp (ifn);
+
+    if (fn == NULL) fp = NULL;
+    else fp = CVS_FOPEN (fn, "w+");
+
+    if (fp == NULL) free (ifn);
+    else chmod (fn, 0600);
+
+    }
+
+#else	/* use tmpnam if all else fails */
+
+    /* tmpnam is deprecated */
+
+    {
+    char ifn[L_tmpnam + 1];
+
+    fn = tmpnam (ifn);
+
+    if (fn == NULL) fp = NULL;
+    else if ((fp = CVS_FOPEN (ifn, "w+")) != NULL)
+    {
+	fn = xstrdup (ifn);
+	chmod (fn, 0600);
+    }
+
+    }
+
+#endif
+
+    *filename = fn;
+    return fp;
 }
 
 /* Return non-zero iff FILENAME is absolute.
@@ -821,6 +932,14 @@ get_homedir ()
     return getenv ("HOME");
 }
 
+#ifndef __VMS_VER
+#define __VMS_VER 0
+#endif
+#ifndef __DECC_VER
+#define __DECC_VER 0
+#endif
+
+#if __VMS_VER < 70200000 || __DECC_VER < 50700000
 /* See cvs.h for description.  On VMS this currently does nothing, although
    I think we should be expanding wildcards here.  */
 void
@@ -834,5 +953,211 @@ expand_wild (argc, argv, pargc, pargv)
     *pargc = argc;
     *pargv = (char **) xmalloc (argc * sizeof (char *));
     for (i = 0; i < argc; ++i)
-	(*pargv)[i] = xstrdup (argv[i]);
+        (*pargv)[i] = xstrdup (argv[i]);
 }
+
+#else  /*  __VMS_VER >= 70200000 && __DECC_VER >= 50700000  */
+
+/* These global variables are necessary to pass information from the
+ * routine that calls decc$from_vms into the callback routine.  In a
+ * multi-threaded environment, access to these variables MUST be
+ * serialized.
+ */
+static char CurWorkingDir[PATH_MAX+1];
+static char **ArgvList;
+static int  CurArg;
+static int  MaxArgs;
+
+static int ew_no_op (char *fname) {
+    (void) fname;   /* Shut the compiler up */
+    return 1;       /* Continue */
+}
+
+static int ew_add_file (char *fname) {
+    char *lastslash, *firstper;
+    int i;
+
+    if (strncmp(fname,CurWorkingDir,strlen(CurWorkingDir)) == 0) {
+        fname += strlen(CurWorkingDir);
+    }
+    lastslash = strrchr(fname,'/');
+    if (!lastslash) {
+        lastslash = fname;
+    }
+    if ((firstper=strchr(lastslash,'.')) != strrchr(lastslash,'.')) {
+        /* We have two periods -- one is to separate the version off */
+        *strrchr(fname,'.') = '\0';
+    }
+    if (firstper && firstper[1]=='\0') {
+        *firstper = '\0';
+    }
+    /* The following code is to insure that no duplicates appear,
+     * because most of the time it will just be a different version
+     */
+    for (i=0;  i<CurArg && strcmp(ArgvList[i],fname)!=0;  ++i) {
+        ;
+    }
+    if (i==CurArg && CurArg<MaxArgs) {
+        ArgvList[CurArg++] = strdup(fname);
+    }
+    return ArgvList[CurArg-1] != 0; /* Stop if we couldn't dup the string */
+}
+
+/* The following two routines are meant to allow future versions of new_arglist
+ * routine to be multi-thread-safe.  It will be necessary in that environment
+ * to serialize access to CurWorkingDir, ArgvList, MaxArg, and CurArg.  We
+ * currently don't do any multi-threaded programming, so right now these
+ * routines are no-ops.
+ */
+static void wait_and_protect_globs (void) {
+    return;
+}
+
+static void release_globs (void) {
+    return;
+}
+
+/*pf---------------------------------------------------------------- expand_wild
+ *
+ *  New Argument List - (SDS)
+ *
+ *  DESCRIPTION:
+ *      This routine takes the argc, argv passed in from main() and returns a
+ *  new argc, argv list, which simulates (to an extent) Unix-Style filename
+ *  globbing with VMS wildcards.  The key difference is that it will return
+ *  Unix-style filenames, i.e., no VMS file version numbers.  The complexity
+ *  comes from the desire to not simply allocate 10000 argv entries.
+ *
+ *  INPUTS:
+ *      argc - The integer argc passed into main
+ *      argv - The pointer to the array of char*'s passed into main
+ *
+ *  OUTPUTS:
+ *      pargv - A pointer to a (char **) to hold the new argv list
+ *      pargc - A pointer to an int to hold the new argc
+ *
+ *  RETURNS:
+ *      NONE
+ *
+ *  SIDE EFFECTS:
+ *      This routine will normally modify the global statics CurArg, MaxArg,
+ *  ArgvList, and CurWorkingDir.
+ *
+ *  NOTES:
+ *      It is ok for &argc == pargc and &argv == pargv.
+ *
+ *------------------------------------------------------------------------------
+ */
+void expand_wild (int argc, char **argv, int *pargc, char ***pargv) {
+    int totfiles, filesgotten;
+    int i;
+    int largc;
+    char **largv;
+
+    /* This first loop is to find out AT MOST how big to make the
+     * pargv array.
+     */
+    for (totfiles=0,i=0;  i<argc;  ++i) {
+        char *arg = argv[i];
+
+        if (arg != 0 && (   strchr(arg,' ') != 0
+                         || strcmp(arg,".") == 0
+                         || strcmp(arg,"..") == 0) ) {
+            ++totfiles;
+        }else if (arg != 0) {
+            int num;
+            char *p = arg;
+            /* Handle comma-separated filelists */
+            while ( (p=strchr(p,',')) != 0) {
+                *p = '\0';
+                num = decc$from_vms (arg, ew_no_op, 1);
+                totfiles += num>0 ? num : 1;
+                *p++ = ',';
+                arg = p;
+            }
+            if (*arg != '\0') {
+                num = decc$from_vms (arg, ew_no_op, 1);
+                totfiles += num>0 ? num : 1;
+            }
+        }
+    }
+    largv = 0;
+    if (totfiles) {
+        largv = malloc (sizeof*largv * (totfiles + 1));
+    }
+    filesgotten = 0;
+    if (largv != 0) {
+        int len;
+        /* All bits set to zero may not be a NULL ptr */
+        for (i=totfiles;  --i>=0;  ) {
+            largv[i] = 0;
+        }
+        largv[totfiles] = 0;
+
+        wait_and_protect_globs ();
+
+        /*--- getcwd has an OpenVMS extension that allows us to ---*/
+        /*--- get back Unix-style path names ---*/
+        (void) getcwd (CurWorkingDir, sizeof CurWorkingDir - 1, 0);
+        len = strlen (CurWorkingDir);
+        if (   len > 0 && CurWorkingDir[len-1] != '/') {
+            (void) strcat (CurWorkingDir, "/");
+        }
+        CurArg = 0;
+        ArgvList = largv;
+        MaxArgs = totfiles + 1;
+
+        for (i=0;  i<argc;  ++i) {
+            char *arg = argv[i];
+
+            if (arg != 0 && (   strchr(arg,' ') != 0
+                             || strcmp(arg,".") == 0
+                             || strcmp(arg,"..") == 0) ) {
+                if (CurArg < MaxArgs) {
+                    ArgvList[CurArg++] = strdup(arg);
+                }
+                ++filesgotten;
+            }else if (arg != 0) {
+                char *p = arg;
+                int num;
+                /* Handle comma-separated filelists */
+                while ( (p=strchr(p,',')) != 0) {
+                    *p = '\0';
+                    num = decc$from_vms (arg, ew_add_file, 1);
+                    if (num <= 0 && CurArg < MaxArgs) {
+                        ArgvList[CurArg++] = strdup(arg);
+                    }
+                    filesgotten += num>0 ? num : 1;
+                    *p++ = ',';
+                    arg = p;
+                }
+                if (*arg != '\0') {
+                    num = decc$from_vms (arg, ew_add_file, 1);
+                    if (num <= 0 && CurArg < MaxArgs) {
+                        ArgvList[CurArg++] = strdup(arg);
+                    }
+                    filesgotten += num>0 ? num : 1;
+                }
+            }
+        }
+        if (filesgotten != totfiles) {
+            /*--- Files must have been created/deleted here ---*/;
+        }
+        filesgotten = CurArg;
+
+        release_globs();
+    }
+    if (!largv) {
+        (*pargv) = malloc (sizeof(char *));
+        if ((*pargv) != 0) {
+            *(*pargv) = 0;
+        }
+    }else {
+        (*pargv) = largv;
+    }
+    (*pargc) = largv ? filesgotten : 0;
+
+    return;
+}
+
+#endif  /*  __VMS_VER >= 70200000 && __DECC_VER >= 50700000  */
