@@ -154,6 +154,9 @@
    this function in the code anyway, hmm.  */
 #undef HAVE_TIMEZONE
 
+/* Define if you have the usleep function.  */
+#define HAVE_USLEEP 1
+
 /* Define if you have the vfork function.  */
 #undef HAVE_VFORK
 
@@ -255,6 +258,8 @@ extern pid_t getpid (void);
 
 /* This is just a call to the Win32 Sleep function.  */
 unsigned int sleep (unsigned int);
+/* So is this */
+int usleep (unsigned long);
 
 /* Don't worry, Microsoft, it's okay for these functions to
    be in our namespace.  */
@@ -291,7 +296,7 @@ extern void wnt_shutdown_server (int fd);
 
 #define SYSTEM_INITIALIZE(pargc,pargv) init_winsock()
 extern void init_winsock();
-#define SYSTEM_CLEANUP() wnt_cleanup
+#define SYSTEM_CLEANUP() wnt_cleanup()
 extern void wnt_cleanup (void);
 
 #define HAVE_WINSOCK_H
@@ -325,3 +330,11 @@ extern char *sock_strerror (int errnum);
    moot since the use of buffer.c ensures that writes will only be as big
    as the buffers).  */
 #define SEND_NEVER_PARTIAL 1
+
+/* Force lib/regex.c to use malloc instead of messing around with alloca
+   and define the old re_comp routines that we use.  */
+#define REGEX_MALLOC 1
+#define _REGEX_RE_COMP 1
+
+/* ssize_t not available under Windows */
+typedef int ssize_t;
